"""
Tools for reactively rendering output for the user interface.
"""

from . import (  # noqa: F401
    transformer,  # pyright: ignore[reportUnusedImport]
)
from ._data_frame import (
    CellPatch,
    CellValue,
    DataGrid,
    DataTable,
    data_frame,
)
from ._data_frame_utils._selection import CellSelection
from ._data_frame_utils._types import (  # noqa: F401
    StyleInfo,
)
from ._deprecated import (  # noqa: F401
    RenderFunction,  # pyright: ignore[reportUnusedImport]
    RenderFunctionAsync,  # pyright: ignore[reportUnusedImport]
)
from ._express import (
    express,
)
from ._render import (
    code,
    download,
    image,
    plot,
    table,
    text,
    ui,
)

__all__ = (
    # TODO-future: Document which variables are exposed via different import approaches
    "data_frame",
    "express",
    "text",
    "code",
    "plot",
    "image",
    "table",
    "ui",
    "download",
    "DataGrid",
    "DataTable",
    "CellPatch",
    "CellValue",
    "CellSelection",
    "StyleInfo",
)



import math
from pathlib import Path

from brownian_motion import brownian_data, brownian_widget
from mediapipe import hand_to_camera_eye, xyz_mean
from shinymediapipe import input_hand
from shinywidgets import output_widget, register_widget

from shiny import App, reactive, render, req, ui

# Check that JS prerequisites are installed
if not (Path(__file__).parent / "shinymediapipe" / "node_modules").is_dir():
    raise RuntimeError(
        "Mediapipe dependencies are not installed. "
        "Please run `npm install` in the 'shinymediapipe' subdirectory."
    )

# Set to True to see underlying XYZ values and canvas
debug = True

app_ui = ui.page_fluid(
    ui.input_action_button("data_btn", "New Data"),
    output_widget("plot"),
    input_hand("hand", debug=debug, throttle_delay_secs=0.05),
    (
        ui.panel_fixed(
            ui.div(ui.tags.strong("x:"), ui.output_text("x_debug", inline=True)),
            ui.div(ui.tags.strong("y:"), ui.output_text("y_debug", inline=True)),
            ui.div(ui.tags.strong("z:"), ui.output_text("z_debug", inline=True)),
            ui.div(ui.tags.strong("mag:"), ui.output_text("mag_debug", inline=True)),
            left="12px",
            bottom="12px",
            width="200px",
            height="auto",
            class_="d-flex flex-column justify-content-end",
        )
        if debug
        else None
    ),
    class_="p-3",
)


def server(input, output, session):
    # BROWNIAN MOTION ====

    @reactive.calc
    def random_walk():
        """Generates brownian data whenever 'New Data' is clicked"""
        input.data_btn()
        return brownian_data(n=200)

    # Create Plotly 3D widget and bind it to output_widget("plot")
    widget = brownian_widget(600, 600)
    register_widget("plot", widget)

    @reactive.effect
    def update_plotly_data():
        walk = random_walk()
        layer = widget.data[0]
        layer.x = walk["x"]
        layer.y = walk["y"]
        layer.z = walk["z"]
        layer.marker.color = walk["z"]

    # HAND TRACKING ====

    @reactive.calc
    def camera_eye():
        """The eye position, as reflected by the hand input"""
        hand_val = input.hand()
        req(hand_val)

        res = hand_to_camera_eye(hand_val, detect_ok=True)
        req(res)
        return res

    # The raw data is a little jittery. Smooth it out by averaging a few samples
    smooth_camera_eye = reactive_smooth(n_samples=5, smoother=xyz_mean)(camera_eye)

    @reactive.effect
    def update_plotly_camera():
        """Update Plotly camera using the hand tracking"""
        widget.layout.scene.camera.eye = smooth_camera_eye()

    # DEBUGGING ====

    @render.text
    def x_debug():
        return camera_eye()["x"]

    @render.text
    def y_debug():
        return camera_eye()["y"]

    @render.text
    def z_debug():
        return camera_eye()["z"]

    @render.text
    def mag_debug():
        eye = camera_eye()

        return f"{round(math.sqrt(eye['x']**2 + eye['y']**2 + eye['z']**2), 2)}"


app = App(app_ui, server)


def reactive_smooth(n_samples, smoother, *, filter_none=True):
    """Decorator for smoothing out reactive calculations over multiple samples"""

    def wrapper(calc):
        buffer = []  # Ring buffer of capacity `n_samples`
        result = reactive.value(None)  # Holds the most recent smoothed result

        @reactive.effect
        def _():
            # Get latest value. Because this is happening in a reactive Effect, we'll
            # automatically take a reactive dependency on whatever is happening in the
            # calc().
            new_value = calc()
            buffer.append(new_value)
            while len(buffer) > n_samples:
                buffer.pop(0)

            if not filter_none:
                result.set(smoother(buffer))
            else:
                # The filter cannot handle None values; remove any in the buffer
                filt_samples = [s for s in buffer if s is not None]
                if len(filt_samples) == 0:
                    result.set(None)
                else:
                    result.set(smoother(filt_samples))

        # The return value for the wrapper
        return result.get

    return wrapper



import logging
import os
import webbrowser

from ._hostenv import get_proxy_url


class LaunchBrowserHandler(logging.Handler):
    """Uvicorn log reader, detects successful app startup so we can launch a browser.

    This class is ONLY used when reload mode is turned off; in the case of reload, the
    launching of the browser must be tightly coupled to the HotReloadHandler as that is
    the only way to ensure the browser is only launched at startup, not on every reload.
    """

    def __init__(self):
        logging.Handler.__init__(self)
        self._launched = False

    def emit(self, record: logging.LogRecord) -> None:
        if self._launched:
            # Ensure that we never launch a browser window twice. In non-reload
            # scenarios it probably would never happen anyway, as we're unlikely to get
            # "Application startup complete." in the logs more than once, but just in
            # case someone does choose to log that string...
            return

        if "Application startup complete." in record.getMessage():
            self._launched = True
            port = os.environ["SHINY_PORT"]
            if not port.isnumeric():
                print(
                    "SHINY_PORT environment variable not set or unusable; "
                    "--launch-browser will be ignored"
                )
                # For some reason the shiny port isn't set correctly!?
                return
            host = os.environ["SHINY_HOST"]
            url = get_proxy_url(f"http://{host}:{port}/")
            webbrowser.open(url, 1)



import pkgutil
import random
import string

import palmerpenguins  # pyright: ignore[reportMissingTypeStubs]
import polars as pl
from htmltools import HTMLDependency

from shiny import App, Inputs, Outputs, Session, render, ui
from shiny.render import CellPatch

pd_penguins = palmerpenguins.load_penguins_raw()
pl_penguins = pl.read_csv(
    pkgutil.get_data(  # pyright: ignore[reportArgumentType]
        "palmerpenguins",
        "data/penguins-raw.csv",
    ),
    null_values="NA",
)


def random_generator():
    return "".join(
        random.choice(string.ascii_letters + string.digits) for _ in range(8)
    )


htmlDep = (
    HTMLDependency(  # pyright: ignore[reportUnknownLambdaType, reportArgumentType]
        f"studyname-0-{random_generator()}",
        version="1",
        head='<script>window.shinytestvalue = "testing"</script>',
    )
)

studyName = (
    pd_penguins["studyName"]  # pyright: ignore[reportUnknownVariableType]
    .copy()
    .astype("object")
)
# Set the first value of the column to an html object so the column is treated as object by narwhals (not str)
studyName[0] = htmlDep
pd_penguins["studyName"] = studyName
pd_penguins["Species"] = pd_penguins["Species"].apply(lambda x: ui.HTML(f"<u>{x}</u>"))  # type: ignore
pd_penguins["Region"] = pd_penguins["Region"].apply(  # type: ignore
    lambda y: ui.tags.h1(  # pyright: ignore[reportUnknownLambdaType, reportArgumentType]
        f"{y}"
    )
)  # pyright: ignore[reportUnknownMemberType]
pd_penguins["Island"] = pd_penguins[
    "Island"
].apply(  # pyright: ignore[reportUnknownMemberType]
    lambda z: ui.TagList(  # pyright: ignore[reportUnknownLambdaType]
        ui.input_checkbox(f"checkbox_{z}_{random_generator()}", f"{z}"),
        ui.tags.img(
            src="https://dka575ofm4ao0.cloudfront.net/pages-transactional_logos/retina/276517/posit-logo-fullcolor-TM.png",
            height="20%",
            width="20%",
        ),
    )
)
# Set the first value of the column to an html object.
# Narwhals does not inspect the types beyond the first row.
# Theforefore, we should not either and need to set the type to object.
pd_penguins.iloc[0, 5] = ui.div(  # pyright: ignore[reportArgumentType]
    pd_penguins.iloc[0, 5]  # pyright: ignore[reportArgumentType]
)
pd_penguins.iloc[1, 5] = ui.p(  # pyright: ignore[reportArgumentType]
    ui.input_action_button("pandas_test_cell_button", "Test button"),
    ui.output_text_verbatim("pandas_test_cell_text", placeholder=True),
)


pl_penguins = (
    pl_penguins.replace_column(
        0,
        pl.Series(
            [
                ui.div(htmlDep) if i == 0 else val
                for i, val in enumerate(pl_penguins["studyName"])
            ]
        ).alias("studyName"),
    )
    .replace_column(
        2,  # Species
        pl.Series([ui.tags.u(x) for x in pl_penguins["Species"].to_list()]).alias(
            "Species"
        ),
    )
    .replace_column(
        3,  # Region
        pl.Series([ui.h1(x) for x in pl_penguins["Region"].to_list()]).alias("Region"),
    )
    .replace_column(
        4,  # Island
        pl.Series(
            [
                ui.div(
                    ui.input_checkbox(f"checkbox_{x}_{random_generator()}", f"{x}"),
                    ui.tags.img(
                        src="https://dka575ofm4ao0.cloudfront.net/pages-transactional_logos/retina/276517/posit-logo-fullcolor-TM.png",
                        height="20%",
                        width="20%",
                    ),
                )
                for x in pl_penguins["Island"]
            ],
        ).alias("Island"),
    )
    .replace_column(
        5,  # Stage
        pl.Series(
            [
                (
                    ui.p(  # pyright: ignore[reportArgumentType]
                        ui.input_action_button(
                            "polars_test_cell_button", "Test button"
                        ),
                        ui.output_text_verbatim(
                            "polars_test_cell_text", placeholder=True
                        ),
                    )
                    if i == 1
                    else ui.div(val)
                )
                for i, val in enumerate(pl_penguins["Stage"])
            ]
        ).alias("Stage"),
    )
    .with_columns(pl.col("Sex").replace("NA", ""))
)


app_ui = ui.page_fluid(
    ui.navset_card_underline(
        ui.nav_panel(
            "init",
            "Click another tab to initialize the larger output data frames",
        ),
        ui.nav_panel(
            "pandas",
            ui.output_data_frame("pandas_df"),
        ),
        ui.nav_panel(
            "polars",
            ui.output_data_frame("polars_df"),
        ),
        id="tab",
        selected="pandas",
    )
)


def server(input: Inputs, output: Outputs, session: Session) -> None:

    @render.text
    def pandas_test_cell_text():
        return f"pandas_test_cell_value {input.pandas_test_cell_button()}"

    @render.text
    def polars_test_cell_text():
        return f"polars_test_cell_value {input.polars_test_cell_button()}"

    @render.data_frame
    def pandas_df():
        return render.DataGrid(
            data=pd_penguins,  # pyright: ignore[reportUnknownArgumentType]
            filters=True,
            editable=True,
        )

    @render.data_frame
    def polars_df():
        return render.DataGrid(
            data=pl_penguins,  # pyright: ignore[reportUnknownArgumentType]
            filters=True,
            editable=True,
        )

    async def upgrade_patch(
        *,
        patch: CellPatch,
    ):
        if patch["column_index"] == 6:  # only for Individual ID
            return f"ID: {patch['value']}"
        else:
            return patch["value"]

    pandas_df.set_patch_fn(upgrade_patch)
    polars_df.set_patch_fn(upgrade_patch)


app = App(app_ui, server)



# This app demonstrates how to use "global" variables that are shared across sessions.
# This is useful if you want to load data just once and use it in multiple apps, or if
# you want to share data or reactives among apps.

import matplotlib.pyplot as plt
import numpy as np
import shared

from shiny import reactive, render
from shiny.express import input, ui


@render.plot
def histogram():
    np.random.seed(19680801)
    x = 100 + 15 * np.random.randn(437)
    plt.hist(x, shared.rv(), density=True)


ui.input_slider("n", "N", 1, 100, 50)


@reactive.effect
def _():
    shared.rv.set(input.n())


@render.code
def rv_value():
    return f"shared.rv() = {shared.rv()}"


@render.code
def text_data():
    return "shared.data = " + str(shared.data)


# If another session changes the slider, make sure the slider also reflects that change.
@reactive.effect
def _():
    ui.update_slider("n", value=shared.rv())



import matplotlib.pyplot as plt
import numpy as np

from shiny.express import input, render, ui

ui.input_slider("obs", "Number of bins:", min=10, max=100, value=30)


@render.plot
def distPlot():
    np.random.seed(19680801)
    x = 100 + 15 * np.random.randn(437)

    fig, ax = plt.subplots()
    ax.hist(x, input.obs(), density=True)
    return fig



from playwright.sync_api import Page
from utils.deploy_utils import skip_if_not_chrome

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@skip_if_not_chrome
def test_patch_exits_saving_state(
    page: Page,
    local_app: ShinyAppProc,
) -> None:
    page.goto(local_app.url)

    my_df = controller.OutputDataFrame(page, "my_df")

    my_df.expect_cell("1", row=0, col=0)
    my_df.expect_cell("4", row=0, col=1)

    my_df.expect_class_state("ready", row=0, col=0)
    my_df.set_cell("test value", row=0, col=1)

    my_df.expect_cell("4", row=0, col=1)
    my_df.expect_cell("a value", row=0, col=0)
    my_df.expect_class_state("success", row=0, col=0)

    my_df.expect_class_state("ready", row=0, col=1)



# -----------------------------------------------------------------------------
# A basic example of dynamically re-rendering a Shiny Chat instance with different models.
# To run it, you'll need an OpenAI API key.
# To get one, follow the instructions at https://platform.openai.com/docs/quickstart
# -----------------------------------------------------------------------------
import os

from langchain_openai import ChatOpenAI

from shiny.express import input, render, ui

ui.input_select("model", "Model", choices=["gpt-4o", "gpt-3.5-turbo"])


@render.express
def chat_ui():

    chat = ui.Chat(
        id="chat",
        messages=[
            {
                "content": f"Hi! I'm a {input.model()} model. How can I help you today?",
                "role": "assistant",
            }
        ],
    )

    chat.ui()

    llm = ChatOpenAI(
        model=input.model(),
        # Provide your API key here (or set the environment variable)
        api_key=os.environ.get("OPENAI_API_KEY"),  # type: ignore
    )

    @chat.on_user_submit
    async def _():
        messages = chat.messages(format="langchain")
        response = llm.astream(messages)
        await chat.append_message_stream(response)



from __future__ import annotations

from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_row_selection(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    df = controller.OutputDataFrame(page, "df1")
    selected_rows = controller.OutputTextVerbatim(page, "selected_rows")
    cell_selection = controller.OutputTextVerbatim(page, "cell_selection")

    df.expect_nrow(3)
    selected_rows.expect_value("Input selected rows: ()")
    cell_selection.expect_value("Cell selection rows: ()")

    df.select_rows([0, 2])

    selected_rows.expect_value("Input selected rows: (0, 2)")
    cell_selection.expect_value("Cell selection rows: (0, 2)")



from datetime import date

from shiny.express import ui

ui.input_date("date1", "Has default date:", value="2016-02-29")
# Default value is the date in client's time zone
ui.input_date("date2", "Client's current date:")
# value is always yyyy-mm-dd, even if the display format is different
ui.input_date("date3", "Format mm/dd/yy:", value="2016-02-29", format="mm/dd/yy")
# Pass in a Date object
ui.input_date("date4", "Default uses date object:", value=date(2016, 2, 29))
# Use different language and different first day of week
ui.input_date(
    "date5",
    "Language is German and the week starts on Monday:",
    language="ru",
    weekstart=1,
)
# Start with decade view instead of default month view
ui.input_date("date6", "Start Date picker in Decade view:", startview="decade")
# Disable Mondays and Tuesdays.
ui.input_date("date7", "Disable Monday and Tuesday:", daysofweekdisabled=[1, 2])
# Disable specific dates.
ui.input_date(
    "date8",
    "Disable specific dates:",
    value="2016-02-29",
    datesdisabled=["2016-03-01", "2016-03-02"],
)
# Set min and max dates.
ui.input_date(
    "date9",
    "Set min and max dates:",
    value="2016-02-03",
    min="2016-02-01",
    max="2016-02-29",
)
# Set width of the date field
ui.input_date("date10", "Set width of text input:", width="600px")
# Set autoclose to false
ui.input_date("date11", "Auto close is disabled:", autoclose=False)



import pytest
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@pytest.mark.parametrize("value_box_id", ["valuebox1", "valuebox2"])
def test_valuebox(page: Page, local_app: ShinyAppProc, value_box_id: str) -> None:
    page.goto(local_app.url)

    value_box = controller.ValueBox(page, value_box_id)
    value_box.expect_full_screen(False)
    value_box.set_full_screen(True)
    value_box.expect_full_screen(True)
    if value_box_id == "valuebox1":
        value_box.expect_height(None)
        value_box.expect_title("KPI Title")
        value_box.expect_value("$1 Billion Dollars")
        value_box.expect_body(["30% VS PREVIOUS 30 DAYS"])
    else:
        value_box.expect_height("500px")
        value_box.expect_title("title")
        value_box.expect_value("value")
        value_box.expect_body(["content", "more body"])
        title_tag_name = (
            value_box.loc_title.locator("*")
            .nth(0)
            .evaluate("el => el.tagName.toLowerCase()")
        )
        assert title_tag_name == "p"
    value_box.set_full_screen(False)
    value_box.expect_full_screen(False)



from shiny import reactive
from shiny.express import input, render, ui

val = reactive.value(0)


@reactive.effect
@reactive.event(input.minus)
def _():
    newVal = val.get() - 1
    val.set(newVal)


@reactive.effect
@reactive.event(input.plus)
def _():
    newVal = val.get() + 1
    val.set(newVal)


with ui.sidebar():
    ui.input_action_button("minus", "-1")
    ui.input_action_button("plus", "+1")


@render.text
def value():
    return str(val.get())



import matplotlib.pyplot as plt
import numpy as np

from shiny import reactive
from shiny.express import input, render, ui

ui.input_slider("n", "Number of observations", min=0, max=1000, value=500)
ui.input_action_link("go", "Go!")


@render.plot(alt="A histogram")
# reactive.event() to invalidate the plot when the button is pressed but not when
# the slider is changed
@reactive.event(input.go, ignore_none=False)
def plot():
    np.random.seed(19680801)
    x = 100 + 15 * np.random.randn(input.n())
    fig, ax = plt.subplots()
    ax.hist(x, bins=30, density=True)
    return fig



from shiny.express import input, render, ui

ui.page_opts(title="Checkbox Group Kitchen Sink", fillable=True)

choices = ["Option A", "Option B", "Option C", "Option D"]

choices_dict = {
    "value1": "Option A",
    "value2": "Option B",
    "value3": "Option C",
    "value4": "Option D",
}

with ui.layout_columns():
    with ui.card():
        ui.card_header("Default Checkbox Group with label")
        ui.input_checkbox_group("default", "Basic Checkbox Group", choices=choices)

        @render.code
        def default_txt():
            return str(input.default())

    with ui.card():
        ui.card_header("With Selected Values")
        ui.input_checkbox_group(
            "selected",
            "Selected Values",
            choices=choices,
            selected=["Option B", "Option C"],
        )

        @render.code
        def selected_txt():
            return str(input.selected())

    with ui.card():
        ui.card_header("With Width")
        ui.input_checkbox_group("width", "Custom Width", choices=choices, width="30px")

        @render.code
        def width_txt():
            return str(input.width())

    with ui.card():
        ui.card_header("Inline")
        ui.input_checkbox_group(
            "inline", "Inline Checkbox Group", choices=choices, inline=True
        )

        @render.code
        def inline_txt():
            return str(input.inline())

    with ui.card():
        ui.card_header("With dict of values")
        ui.input_checkbox_group(
            "dict_values",
            "Dict Values",
            choices=choices_dict,
        )

        @render.code
        def dict_values_txt():
            return str(input.dict_values())



from shiny import reactive
from shiny.express import render, ui

chat = ui.Chat(id="chat")
chat.ui()


@reactive.effect
async def _():
    await chat.append_message({"content": "A user message", "role": "user"})


"chat.messages():"


@render.code
def message_state():
    return str(chat.messages())



from __future__ import annotations

import re

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect
from typing_extensions import Literal

from shiny.types import ListOrTuple

from .._types import PatternOrStr, Timeout
from ..expect import expect_to_have_class, expect_to_have_style
from ..expect._internal import expect_attribute_to_have_value
from ..expect._internal import expect_class_to_have_value as _expect_class_to_have_value
from ._base import (
    InitLocator,
    UiWithContainer,
    UiWithContainerP,
    UiWithSidebarP,
    UiWithTitleP,
)
from ._expect import expect_locator_values_in_list


class _ExpectNavsetSidebarM:
    def expect_sidebar(
        self: UiWithSidebarP,
        exists: bool,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Assert whether or not the sidebar exists within the navset.

        Parameters
        ----------
        exists
            `True` if the sidebar exists within the navset.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc_sidebar).to_have_count(int(exists), timeout=timeout)


class _ExpectNavsetTitleM:
    """A mixin class for Navset title controls"""

    def expect_title(
        self: UiWithTitleP,
        value: PatternOrStr,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the navset title to have the specified text.

        Parameters
        ----------
        value
            The expected text pattern or string.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc_title).to_have_text(value, timeout=timeout)


class _ExpectNavsetPlacementM:
    def expect_placement(
        self: UiWithContainerP,
        location: Literal["above", "below"] = "above",
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the navset to have the specified placement.

        Parameters
        ----------
        location
            The expected placement location. Defaults to `'above'`.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        ex_class = "card-header" if location == "above" else "card-footer"
        expect_to_have_class(
            self.loc_container.locator(".."), ex_class, timeout=timeout
        )


class NavPanel(UiWithContainer):
    """Controller for :func:`shiny.ui.nav_panel`."""

    """
    Playwright `Locator` for the content of the nav panel.
    """
    loc: Locator
    """
    Playwright `Locator` for the nav panel.
    """
    loc_container: Locator
    """
    Playwright `Locator` for the nav panel container.
    """
    panel_value: str
    """
    The `data-value` attribute used to identify the nav panel within the larger navset.
    """

    def __init__(self, page: Page, id: str, panel_value: str) -> None:
        """
        Initializes a new instance of the `NavPanel` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the nav panel.
        panel_value
            The panel value of the nav panel.
        """
        super().__init__(
            page,
            id=id,
            loc=f"a[role='tab'][data-value='{panel_value}']",
            loc_container=f"ul#{id}",
        )

        self.panel_value: str = panel_value

    # TODO-future: Make it a single locator expectation
    # get active content instead of assertion
    @property
    def loc_content(self) -> Locator:
        """
        Returns the locator for the content of the nav panel.

        Note: This requires 2 steps. Will not work if the overlay element is rapidly created during locator fetch
        """
        datatab_id = self.loc_container.get_attribute("data-tabsetid")
        return self.page.locator(
            f"div.tab-content[data-tabsetid='{datatab_id}'] > div.tab-pane[data-value='{self.panel_value}']"
        )

    def click(self, *, timeout: Timeout = None) -> None:
        """
        Clicks the nav panel.

        Parameters
        ----------
        timeout
            The maximum time to wait for the nav panel to be visible and interactable. Defaults to `None`.
        """
        self.loc.click(timeout=timeout)

    def expect_active(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expects the nav panel to be active or inactive.

        Parameters
        ----------
        value
            `True` if the nav panel is expected to be active, False otherwise.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self.loc,
            "active",
            has_class=value,
            timeout=timeout,
        )

    def _expect_content_text(
        self, value: PatternOrStr, *, timeout: Timeout = None
    ) -> None:
        """
        Expects the nav panel content to have the specified text.

        Parameters
        ----------
        value
            The expected text pattern or string.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.loc_content).to_have_text(value, timeout=timeout)


class _NavsetBase(UiWithContainer):
    """A Base mixin class for Nav controls"""

    def nav_panel(
        self,
        value: str,
    ) -> NavPanel:
        """
        Returns the nav panel (:class:`~shiny.playwright.controls.NavPanel`)
        with the specified value.

        Parameters
        ----------
        value
            The value of the nav panel.
        """
        return NavPanel(self.page, self.id, value)

    def set(self, value: str, *, timeout: Timeout = None) -> None:
        """
        Sets the state of the control to open or closed.

        Parameters
        ----------
        value
            The selected nav item.
        """
        self.nav_panel(value).click(timeout=timeout)

    def expect_value(self, value: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Expects the control to have the specified value.

        Parameters
        ----------
        value
            The expected value.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        # data attribute of active tab and compare with value
        playwright_expect(
            self.loc_container.locator('a[role="tab"].active')
        ).to_have_attribute("data-value", value, timeout=timeout)

    # # TODO-future: Make it a single locator expectation
    # # get active content instead of assertion
    # @property
    # def loc_active_content(self) -> Locator:
    #     datatab_id = self.loc_container.get_attribute("data-tabsetid")
    #     return self.page.locator(
    #         f"div.tab-content[data-tabsetid='{datatab_id}'] > div.tab-pane.active"
    #     )

    def get_loc_active_content(self, *, timeout: Timeout = None) -> Locator:
        """
        Returns the locator for the active content.

        Parameters
        ----------
        timeout
            The maximum time to wait for the locator to appear. Defaults to `None`.
        """
        datatab_id = self.loc_container.get_attribute("data-tabsetid", timeout=timeout)
        return self.page.locator(
            f"div.tab-content[data-tabsetid='{datatab_id}'] > div.tab-pane.active"
        )

    def _expect_content_text(
        self, value: PatternOrStr, *, timeout: Timeout = None
    ) -> None:
        """
        Expects the control to have the specified content.

        Parameters
        ----------
        value
            The expected content.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.get_loc_active_content()).to_have_text(
            value, timeout=timeout
        )

    def expect_nav_values(
        self,
        value: list[PatternOrStr],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the control to have the specified nav values.

        Parameters
        ----------
        value
            The expected nav values.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        expect_locator_values_in_list(
            page=self.page,
            loc_container=self.loc_container,
            el_type="a[role='tab']",
            arr_name="value",
            arr=value,
            key="data-value",
            timeout=timeout,
        )

    def expect_nav_titles(
        self,
        value: list[PatternOrStr],
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the control to have the specified nav titles.

        Parameters
        ----------
        value
            The expected nav titles.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        self.expect.to_have_text(value, timeout=timeout)

    # # 2024-08-23-barret:
    # # These two functions are not implemented due to the inability to create a locator
    # # as some of the text contents are not contained within a defined parent element.
    # # This makes querying the header only (or footer only) impossible. When they are
    # # used within `navset_card_*()`, the header or footer _could_ be given within a
    # # `core_ui.CardItem()`. This CardItem could contain a TagList, putting us back into
    # # the same situation. Therefore, no move is currently the safe move. If we want to
    # # expose anything, maybe we could expose the navset card body container, but there
    # # is no container for the non-card navsets. :-(
    # def expect_header():
    #     raise NotImplementedError("Not implemented yet")
    # def expect_footer():
    #     raise NotImplementedError("Not implemented yet")


class NavsetTab(_NavsetBase):
    """Controller for :func:`shiny.ui.navset_tab`."""

    loc: Locator
    """
    Playwright `Locator` for the nav set tab.
    """
    loc_container: Locator
    """
    Playwright `Locator` for the nav set tab container.
    """

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `NavsetTab` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the nav set tab.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f"ul#{id}.nav-tabs",
            loc="a[role='tab']",
        )


class NavsetPill(_NavsetBase):
    """Controller for :func:`shiny.ui.navset_pill`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `NavsetPill` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the nav set pill.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f"ul#{id}.nav-pills",
            loc="> li.nav-item",
        )


class NavsetUnderline(_NavsetBase):
    """Controller for :func:`shiny.ui.navset_underline`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `NavsetUnderline` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the nav set underline.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f"ul#{id}.nav-underline",
            loc="> li.nav-item",
        )


class NavsetPillList(_NavsetBase):
    """Controller for :func:`shiny.ui.navset_pill_list`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `NavsetPillList` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the nav set pill list.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f"ul#{id}.nav-stacked",
            loc="> li.nav-item",
        )

    def expect_well(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expects the navset pill list to have a well.

        Parameters
        ----------
        value
            `True` if the navset pill list is expected to be constructed with a well,
            `False` otherwise.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self.loc_container.locator(".."), "well", has_class=value, timeout=timeout
        )

    def expect_widths(
        self, value: ListOrTuple[int], *, timeout: Timeout = None
    ) -> None:
        """
        Expects the navset pill list to have the specified widths.

        Parameters
        ----------
        value
            The expected widths of the navset pill list.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        widths = tuple(value)
        assert len(widths) == 2, "`value=` must be a tuple of two integers"
        assert all(
            isinstance(width, int) for width in widths
        ), "`value=` must be integers"

        loc_row_container = self.loc_container.locator("..").locator("..")

        # Make sure the two children are present
        loc_complicated = loc_row_container.locator(
            "xpath=.",
            has=self.page.locator(f"> div.col-sm-{widths[0]} + div.col-sm-{widths[1]}"),
        )

        # Make sure there are only two children present
        try:
            playwright_expect(loc_complicated.locator("> div")).to_have_count(
                2, timeout=timeout
            )
        except AssertionError as e:
            # Make sure there are only two children
            playwright_expect(loc_row_container.locator("> div")).to_have_count(
                2, timeout=1
            )

            expect_to_have_class(
                loc_row_container.locator("> div").first,
                f"col-sm-{widths[0]}",
                timeout=1,
            )
            expect_to_have_class(
                loc_row_container.locator("> div").last,
                f"col-sm-{widths[1]}",
                timeout=1,
            )

            # Re-raise the original exception if nothing could be debugged
            raise e


class _NavsetCardBase(
    _ExpectNavsetSidebarM,
    _ExpectNavsetTitleM,
    _NavsetBase,
):
    def __init__(
        self,
        page: Page,
        *,
        id: str,
        loc: InitLocator,
        loc_container: InitLocator,
    ) -> None:
        """
        Shim class to add consistent `.loc_sidebar` definition
        """
        super().__init__(
            page,
            id=id,
            loc_container=loc_container,
            loc=loc,
        )
        self.loc_sidebar = self.loc_container.locator("..").locator(
            "+ .bslib-sidebar-layout"
        )
        self.loc_title = (
            self.loc_container.locator("..")
            .locator("> span")
            .locator("xpath=.", has=self.page.locator(f"+ ul#{self.id}"))
        )


class NavsetCardTab(_NavsetCardBase):
    """Controller for :func:`shiny.ui.navset_card_tab`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `NavsetCardTab` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the nav set card tab.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f".bslib-card > div > ul#{id}.nav-tabs",
            loc="> li.nav-item",
        )


class NavsetCardPill(_ExpectNavsetPlacementM, _NavsetCardBase):
    """Controller for :func:`shiny.ui.navset_card_pill`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `NavsetCardPill` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the nav set card pill.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f".bslib-card > div > ul#{id}.nav-pills",
            loc="> li.nav-item",
        )


class NavsetCardUnderline(_ExpectNavsetPlacementM, _NavsetCardBase):
    """Controller for :func:`shiny.ui.navset_card_underline`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `NavsetCardUnderline` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the nav set card underline.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f".bslib-card > div > ul#{id}.nav-underline",
            loc="> li.nav-item",
        )


class NavsetHidden(_NavsetBase):
    """Controller for :func:`shiny.ui.navset_hidden`."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `NavsetHidden` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the nav set hidden.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f"ul#{id}.nav-hidden",
            loc="> li.nav-item",
        )


class _NavsetBarBase(
    _ExpectNavsetSidebarM,
    _ExpectNavsetTitleM,
    _NavsetBase,
):
    """Mixin class for common expectations of nav bars."""

    def __init__(self, page: Page, id: str) -> None:
        """
        Initializes a new instance of the `NavsetBar` class.

        Parameters
        ----------
        page
            Playwright `Page` of the Shiny app.
        id
            The ID of the nav set bar.
        """
        super().__init__(
            page,
            id=id,
            loc_container=f"ul#{id}.navbar-nav",
            loc="> li.nav-item",
        )
        self._loc_navbar = self.loc_container.locator("..").locator("..").locator("..")

        # This location is different than the `_NavsetCardBase.loc_title`
        self.loc_title = (
            self.loc_container.locator("..").locator("..").locator("> .navbar-brand")
        )
        # This location is different than the `_NavsetCardBase.loc_sidebar`
        self.loc_sidebar = (
            self.loc_container.locator("..")
            .locator("..")
            .locator("..")
            .locator("+ div > .bslib-sidebar-layout")
        )

    def expect_position(
        self,
        position: Literal[
            "fixed-top", "fixed-bottom", "static-top", "sticky-top"
        ] = "static-top",
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the navset bar to have the specified position.

        Parameters
        ----------
        position
            The expected position. Defaults to `'static-top'`.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        if position == "static-top":
            # static-top class is not provided by the code.
            # Therefore we must check that all other classes are **not** found
            playwright_expect(self._loc_navbar).not_to_have_class(
                re.compile(r"(^|\s+)(fixed-top|fixed-bottom|sticky-top)(\s+|$)"),
                timeout=timeout,
            )
        else:
            expect_to_have_class(self._loc_navbar, position, timeout=timeout)

    def expect_inverse(
        self,
        value: bool,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the navset bar to be light text color if inverse is True

        Parameters
        ----------
        value
            `True` if the navset bar is expected to have inverse text color, `False` otherwise.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        _expect_class_to_have_value(
            self._loc_navbar,
            "navbar-inverse",
            has_class=value,
            timeout=timeout,
        )

    def expect_bg(self, bg: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Expects the navset bar to have the specified background color.

        Parameters
        ----------
        bg
            The expected background color.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        expect_to_have_style(
            self._loc_navbar,
            "background-color",
            f"{bg} !important",
            timeout=timeout,
        )

    def expect_gap(self, gap: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Expects the navset bar to have the specified gap.

        Parameters
        ----------
        gap
            The expected gap.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        expect_to_have_style(
            self.get_loc_active_content(),
            "gap",
            gap,
            timeout=timeout,
        )

    def expect_fluid(
        self,
        value: bool,
        *,
        timeout: Timeout = None,
    ) -> None:
        """
        Expects the navset bar to have a fluid or fixed layout.

        Parameters
        ----------
        value
            `True` if the layout is `fluid` or `False` if it is `fixed`.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        if value:
            expect_to_have_class(
                self._loc_navbar.locator("> div"),
                "container-fluid",
                timeout=timeout,
            )
        else:
            expect_to_have_class(
                self._loc_navbar.locator("> div"),
                "container",
                timeout=timeout,
            )


class NavsetBar(_NavsetBarBase):
    """Controller for :func:`shiny.ui.navset_bar`."""


class PageNavbar(_NavsetBarBase):
    """Controller for :func:`shiny.ui.page_navbar`."""

    def expect_fillable(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expects the main content area to be considered a fillable (i.e., flexbox) container

        Parameters
        ----------
        value
            `True` if the main content area is expected to be fillable, `False` otherwise.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        # confirm page is fillable
        _expect_class_to_have_value(
            self.page.locator("body"),
            "bslib-page-fill",
            has_class=value,
            timeout=timeout,
        )

        # confirm content is fillable
        _expect_class_to_have_value(
            self.get_loc_active_content(),
            "html-fill-container",
            has_class=value,
            timeout=timeout,
        )

    def expect_fillable_mobile(self, value: bool, *, timeout: Timeout = None) -> None:
        """
        Expects the main content area to be considered a fillable (i.e., flexbox) container on mobile
        This method will always call `.expect_fillable(True)` first to ensure the fillable property is set

        Parameters
        ----------
        value
            `True` if the main content area is expected to be fillable on mobile, `False` otherwise.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """

        # This is important since fillable_mobile needs fillable property to be True
        self.expect_fillable(True, timeout=timeout)
        _expect_class_to_have_value(
            self.page.locator("body"),
            "bslib-flow-mobile",
            has_class=not value,
            timeout=timeout,
        )

    def expect_window_title(
        self, title: PatternOrStr, *, timeout: Timeout = None
    ) -> None:
        """
        Expects the window title to have the specified text.

        Parameters
        ----------
        title
            The expected window title.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        playwright_expect(self.page).to_have_title(title, timeout=timeout)

    def expect_lang(self, lang: PatternOrStr, *, timeout: Timeout = None) -> None:
        """
        Expects the HTML tag to have the specified language.

        Parameters
        ----------
        lang
            The expected language.
        timeout
            The maximum time to wait for the expectation to pass. Defaults to `None`.
        """
        expect_attribute_to_have_value(
            self.page.locator("html"),
            "lang",
            lang,
            timeout=timeout,
        )



import pytest
from examples.example_apps import reruns, reruns_delay
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_accordion(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    acc = controller.Accordion(page, "acc")
    acc_panel_A = acc.accordion_panel("Section A")
    output_txt_verbatim = controller.OutputTextVerbatim(page, "acc_txt")
    alternate_button = controller.InputActionButton(page, "alternate")
    open_all_button = controller.InputActionButton(page, "open_all")
    close_all_button = controller.InputActionButton(page, "close_all")
    toggle_b_button = controller.InputActionButton(page, "toggle_b")
    toggle_updates_button = controller.InputActionButton(page, "toggle_updates")
    toggle_efg_button = controller.InputActionButton(page, "toggle_efg")

    output_txt_verbatim.expect_value("input.acc(): ('Section A',)")

    acc_panel_A.set(True)
    acc_panel_A.expect_open(True)
    acc_panel_A.set(False)
    acc_panel_A.expect_open(False)

    acc.expect_width(None)
    acc.expect_height(None)

    close_all_button.click()
    acc.expect_open([])
    output_txt_verbatim.expect_value("input.acc(): None")
    acc.set(["Section A"])

    acc.expect_panels(["Section A", "Section B", "Section C", "Section D"])
    output_txt_verbatim.expect_value("input.acc(): ('Section A',)")
    acc.expect_open(["Section A"])
    acc_panel_A.expect_label("Section A")
    acc_panel_A.expect_body("Some narrative for section A")
    acc_panel_A.expect_open(True)

    alternate_button.click()
    acc.expect_open(["Section B", "Section D"])
    output_txt_verbatim.expect_value("input.acc(): ('Section B', 'Section D')")

    alternate_button.click()
    acc.expect_open(["Section A", "Section C"])
    output_txt_verbatim.expect_value("input.acc(): ('Section A', 'Section C')")

    open_all_button.click()
    acc.expect_open(["Section A", "Section B", "Section C", "Section D"])
    output_txt_verbatim.expect_value(
        "input.acc(): ('Section A', 'Section B', 'Section C', 'Section D')"
    )

    toggle_b_button.click()
    acc.expect_open(["Section A", "Section C", "Section D"])
    output_txt_verbatim.expect_value(
        "input.acc(): ('Section A', 'Section C', 'Section D')"
    )

    acc_panel_updated_A = acc.accordion_panel("updated_section_a")
    toggle_updates_button.click()
    acc_panel_updated_A.expect_label("Updated title")
    acc_panel_updated_A.expect_body("Updated body")
    acc_panel_updated_A.expect_icon(True)

    acc.expect_panels(["updated_section_a", "Section B", "Section C", "Section D"])
    # workaround - toggle it twice Section A
    acc_panel_updated_A._toggle()
    # add timeout to wait for css animation
    page.wait_for_timeout(100)
    acc_panel_updated_A._toggle()

    output_txt_verbatim.expect_value(
        "input.acc(): ('updated_section_a', 'Section C', 'Section D')"
    )

    toggle_efg_button.click()
    acc.expect_panels(
        [
            "updated_section_a",
            "Section B",
            "Section C",
            "Section D",
            "Section E",
            "Section F",
            "Section G",
        ],
    )
    acc.expect_open(
        [
            "updated_section_a",
            "Section C",
            "Section D",
            "Section E",
            "Section F",
            "Section G",
        ],
    )
    # Should be uncommented once https://github.com/rstudio/bslib/issues/565 is fixed
    # output_txt_verbatim.expect_value(
    #     "input.acc(): ('updated_section_a', 'Section B', 'Section E', 'Section F', 'Section G')"
    # )



from shiny import reactive
from shiny.express import ui

ui.input_selectize("x", "Server side selectize", choices=[], multiple=True)


@reactive.effect
def _():
    ui.update_selectize(
        "x",
        choices=[f"Foo {i}" for i in range(10000)],
        selected=["Foo 0", "Foo 1"],
        server=True,
    )



from __future__ import annotations

from typing import Iterable, Literal, Optional, cast

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, css, div

from .._deprecated import warn_deprecated
from .._docstring import add_example
from ..types import MISSING, MISSING_TYPE
from ._html_deps_shinyverse import components_dependencies
from ._tag import consolidate_attrs
from ._utils import is_01_scalar
from .css import CssUnit, as_css_unit
from .css._css_unit import isinstance_cssunit
from .fill import as_fill_item, as_fillable_container


@add_example()
def layout_column_wrap(
    *args: TagChild | TagAttrs,
    width: CssUnit | None | MISSING_TYPE = MISSING,
    fixed_width: bool = False,
    heights_equal: Literal["all", "row"] = "all",
    fill: bool = True,
    fillable: bool = True,
    height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    height_mobile: Optional[CssUnit] = None,
    gap: Optional[CssUnit] = None,
    class_: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    # For
    # more explanation and illustrative examples, see
    # [here](https://rstudio.github.io/bslib/articles/cards.html#multiple-cards)
    """
    A grid-like, column-first layout

    Wraps a 1d sequence of UI elements into a 2d grid. The number of columns
    (and rows) in the grid depends on the column width and the size of the display.

    Parameters
    ----------
    *args
        Unnamed arguments should be UI elements (e.g.,
        :func:`~shiny.ui.card`). Named arguments become attributes on the
        containing :class:`~htmltools.Tag` element.
    width
        The desired width of each card. It can be one of the following:

        * A (unit-less) number between 0 and 1, specified as `1/num`, where `num`
          represents the number of desired columns.
        * A CSS length unit representing either the minimum (when `fixed_width=False`)
          or fixed width (`fixed_width=True`).
        * `None`, which allows power users to set the `grid-template-columns` CSS
          property manually, either via a `style` attribute or a CSS stylesheet.
        * If missing, a value of `200px` will be used.
    fixed_width
        When `width` is greater than 1 or is a CSS length unit, e.g., `"200px"`,
        `fixed_width` indicates whether that `width` value represents the absolute size
        of each column (`fixed_width=TRUE`) or the minimum size of a column
        (`fixed_width=FALSE`).

        When `fixed_width=FALSE`, new columns are added to a row when `width` space is
        available and columns will never exceed the container or viewport size.

        When `fixed_width=TRUE`, all columns will be exactly `width` wide, which may
        result in columns overflowing the parent container.
    heights_equal
        If `"all"` (the default), every card in every row of the grid will have the same
        height. If `"row"`, then every card in _each_ row of the grid will have the same
        height, but heights may vary between rows.
    fill
        Whether or not to allow the layout to grow/shrink to fit a fillable container
        with an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).
    fillable
        Whether or not each element is wrapped in a fillable container.
    height,max_height,min_height
        A valid CSS unit (e.g., `height="200px"`). Use `min_height` and `max_height` in
        a filling layout to ensure that the layout container does not shrink below a
        `min_height` or grow beyond a `max_height`.
    height_mobile
        Any valid CSS unit to use for the height when on mobile devices (or narrow
        windows).
    gap
        Any valid CSS unit to use for the gap between columns.
    class_
        A CSS class to apply to the containing element.
    **kwargs
        Additional attributes to apply to the containing element.

    Returns
    -------
    :
        A :class:`~htmltools.Tag` element.

    See Also
    --------
    * :func:`~shiny.ui.layout_columns` for laying out elements into a responsive
      12-column grid.
    """
    attrs, children = consolidate_attrs(*args, class_=class_, **kwargs)

    # TODO-future; Remove this when we remove `shiny.experimental.ui.layout_column_wrap()`
    if isinstance(width, MISSING_TYPE):
        if len(children) > 0 and (
            children[0] is None or is_probably_a_css_unit(children[0])
        ):
            # Code changed from
            #   `layout_column_wrap(width, *args)`
            # to
            #   `layout_column_wrap(*args, width)`.
            # Provide deprecation warning

            # Assume an unnamed first argument that matches our expectations for
            # `width` is actually the width argument, with a warning
            warn_deprecated(
                "`layout_column_wrap(*args, width=)`'s `width` parameter must be named."
            )
            width = cast(CssUnit, children[0])
            children = children[1:]
        else:
            width = "200px"

    colspec: str | None = None
    if width is not None:
        if is_01_scalar(width) and width > 0.0:
            num_cols = 1.0 / width
            if not num_cols.is_integer():
                raise ValueError(
                    "Could not interpret `layout_column_wrap(width=)` argument"
                )
            colspec = " ".join(["1fr" for _ in range(int(num_cols))])
        else:
            width_css_unit = as_css_unit(width)
            if fixed_width:
                colspec = f"repeat(auto-fit, minmax({width_css_unit}, 1fr))"
            else:
                colspec = f"repeat(auto-fit, minmax(min({width_css_unit}, 100%), 1fr))"

    tag_style_css = {
        "grid-template-columns": colspec,
        "grid-auto-rows": "1fr" if (heights_equal == "all") else None,
        # Always provide the `height:auto` default so that the CSS variable
        # doesn't get inherited in a scenario like
        # layout_column_wrap(height=200, layout, layout_column_wrap(...))
        "--bslib-grid-height": as_css_unit("auto" if height is None else height),
        "--bslib-grid-height-mobile": as_css_unit(
            "auto" if height_mobile is None else height_mobile
        ),
        "gap": as_css_unit(gap),
        "min-height": as_css_unit(min_height),
        "max-height": as_css_unit(max_height),
    }

    tag = div(
        {
            "class": "bslib-grid bslib-mb-spacing",
            "style": css(**tag_style_css),
        },
        attrs,
        *wrap_all_in_gap_spaced_container(children, fillable),
        components_dependencies(),
    )
    if fill:
        tag = as_fill_item(tag)

    return tag


def is_probably_a_css_unit(x: TagChild) -> bool:
    if isinstance(x, str):
        return False
    if isinstance_cssunit(x):
        return True
    return False


def wrap_all_in_gap_spaced_container(
    children: Iterable[TagChild],
    fillable: bool = True,
    class_: Optional[str] = None,
) -> list[TagChild]:
    item_class = "bslib-gap-spacing"
    if class_ is not None:
        item_class = f"{item_class} {class_}"

    # Use a new list so that we don't mutate the original `children`
    wrapped_children: list[TagChild] = []
    for child_value in children:
        child = div({"class": item_class}, child_value)
        if fillable:
            child = as_fillable_container(child)
        wrapped_children.append(child)

    return wrapped_children



import pandas as pd
import polars as pl
import seaborn as sns

from shiny import App, Inputs, Outputs, Session, render, ui

df = pd.DataFrame(
    sns.load_dataset(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
        "iris"
    )
)
pl_df = pl.from_pandas(df)
df["sepal_length"] = df["sepal_length"].apply(lambda x: ui.tags.u(x))  # pyright: ignore
df["sepal_width"] = df["sepal_width"].apply(lambda x: ui.tags.u(x))  # pyright: ignore
df["petal_width"] = df["petal_width"].apply(lambda x: ui.tags.u(x))  # pyright: ignore
df["species"] = df["species"].apply(lambda x: ui.tags.u(x))  # pyright: ignore

pl_df = pl_df.with_columns(
    (pl.Series([ui.tags.u(x) for x in pl_df["sepal_length"].to_list()])).alias(
        "sepal_length"
    ),
    (pl.Series([ui.tags.u(x) for x in pl_df["sepal_width"].to_list()])).alias(
        "sepal_width"
    ),
    (pl.Series([ui.tags.u(x) for x in pl_df["petal_width"].to_list()])).alias(
        "petal_width"
    ),
    (pl.Series([ui.tags.u(x) for x in pl_df["species"].to_list()])).alias("species"),
)

app_ui = ui.page_fluid(
    ui.navset_card_underline(
        ui.nav_panel("pandas", ui.output_data_frame("pandas_iris")),
        ui.nav_panel("polars", ui.output_data_frame("polars_iris")),
        id="tab",
    ),
)


def server(input: Inputs, output: Outputs, session: Session) -> None:

    @render.data_frame
    def pandas_iris():
        return render.DataGrid(
            data=df.head(),  # pyright: ignore[reportUnknownArgumentType]
            editable=True,
        )

    @render.data_frame
    def polars_iris():
        return render.DataGrid(
            data=pl_df.head(),  # pyright: ignore[reportUnknownArgumentType]
            editable=True,
        )


app = App(app_ui, server)



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_select")


def test_input_select_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    state = controller.InputSelect(page, "state")

    expect(state.loc_label).to_have_text("Choose a state:")
    state.expect_label("Choose a state:")

    state.expect_choices(["NY", "NJ", "CT", "WA", "OR", "CA", "MN", "WI", "IA"])
    state.expect_choice_labels(
        [
            "New York",
            "New Jersey",
            "Connecticut",
            "Washington",
            "Oregon",
            "California",
            "Minnesota",
            "Wisconsin",
            "Iowa",
        ]
    )
    state.expect_choice_groups(["East Coast", "West Coast", "Midwest"])

    state.expect_selected("NY")
    state.expect_multiple(False)
    # state.expect_selectize(False)

    state.expect_width(None)

    state.set("IA")

    state.expect_selected("IA")



import pytest
from playwright.sync_api import Page
from utils.deploy_utils import skip_if_not_chrome

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@skip_if_not_chrome
@pytest.mark.parametrize("df_type", ["pandas", "polars"])
def test_edit_cell_content_is_not_null(
    page: Page, local_app: ShinyAppProc, df_type: str
) -> None:
    page.goto(local_app.url)

    data_frame = controller.OutputDataFrame(page, f"{df_type}_df")

    tab = controller.NavsetCardUnderline(page, "tab")
    tab.set(df_type)

    data_frame.expect_cell("", row=0, col=14)
    empty_cell = data_frame.cell_locator(row=0, col=14)
    empty_cell.dblclick()
    textarea = empty_cell.locator("textarea")
    textarea.wait_for()
    cur_value = textarea.input_value()
    assert cur_value == ""



from __future__ import annotations

import datetime
import os
import pathlib
import sys
import typing
from typing import Any, Callable

import matplotlib.pyplot as plt

# =============================================================================
# Data
# =============================================================================
import pandas as pd
from htmltools import TagList

from shiny import App, Inputs, Outputs, Session, module, reactive, render, ui
from shiny.session import session_context
from shiny.types import ImgData

pandas_df = pd.DataFrame(
    {
        "name": ["Mazda RX4", "Mazda RX4 Wag", "Datsun 710", "Hornet 4 Drive"],
        "mpg": [21.0, 21.0, 22.8, 21.4],
        "cyl": [6, 6, 4, 6],
    }
)

# To add more options, add more images to the `imgs` folder!!
img_path = pathlib.Path(__file__).parent / "imgs"
penguin_imgs = [str(img_path / img) for img in os.listdir(img_path)]
assert len(penguin_imgs) > 0
letters = [letter for letter in "abcdefghijklmnopqrstuvwxyz"][: len(penguin_imgs)]

input_keys = (
    "input_action_button",
    "input_action_link",
    "download_button",
    "download_link",
    "accordion",
    "input_file",
    "input_checkbox",
    "input_checkbox_group",
    "input_date",
    "input_date_range",
    "input_numeric",
    "input_password",
    "input_radio_buttons",
    "input_select",
    "input_selectize",
    "input_slider",
    "input_switch",
    "input_text",
    "input_text_area",
    "navset_bar",
    "navset_card_pill",
    "navset_card_tab",
    "navset_hidden",
    "navset_pill",
    "navset_tab",
    "sidebar",
    "popover",
    "tooltip",
)
x_input_keys = ()


session_dict: dict[str, Session] = {}


# ============================================================
# Counter module
# ============================================================
@module.ui
def mod_x_ui(label: str) -> ui.TagChild:
    def ui_navs(label: str) -> list[ui._navs.NavPanel]:
        return [
            ui.nav_panel(
                letter.capitalize(),
                f"{label} - {letter.capitalize()} content",
                value=letter,
            )
            for letter in letters
        ]

    return ui.layout_sidebar(
        ui.sidebar("Sidebar content", id="sidebar", width="100px"),
        ui.h2(label),
        ui.tags.label("Output status:"),
        ui.tags.style(
            """
            pre[id*="status_"] {
                padding: 0;
                font-size: xx-small;
                text-wrap: balance;
            }
            """
        ),
        ui.layout_column_wrap(
            *[
                ui.output_text_verbatim(f"status_x_{x_input_key}", placeholder=True)
                for x_input_key in x_input_keys
            ],
            *[
                ui.output_text_verbatim(f"status_{input_key}", placeholder=True)
                for input_key in input_keys
            ],
            width=1 / 2,
            gap="2px",
            heights_equal="row",
        ),
        ui.tags.label("Inputs that need to be visible when updated:"),
        ui.card(
            ui.popover(
                ui.input_action_button("popover_btn", "Popover btn"),
                "Popover content",
                id="popover",
            ),
            ui.tooltip(
                ui.input_action_button("tooltip_btn", "Tooltip btn"),
                "Tooltip content",
                id="tooltip",
            ),
        ),
        ui.tags.label("Manual inputs / outputs:"),
        ui.card(
            ui.input_action_button("input_action_button", "input_button"),
            ui.input_action_link("input_action_link", "Link"),
            ui.download_button("download_button", "Download"),
            ui.download_link("download_link", "Download Link"),
            ui.input_file("input_file", "File"),
            # Plot
            ui.tags.label(ui.markdown("Plot (Changes w/ `input.radio_buttons()`)")),
            ui.output_plot("out_plot"),
        ),
        ui.accordion(
            *[
                ui.accordion_panel(
                    ui.markdown(f"Panel for `{letter}`"),
                    ui.markdown(f"Content for `{letter}`"),
                    value=letter,
                )
                for letter in letters
            ],
            id="accordion",
        ),
        ui.input_checkbox("input_checkbox", "Checkbox"),
        ui.input_checkbox_group(
            "input_checkbox_group", "Checkbox group", choices=letters
        ),
        ui.input_date("input_date", "Date", value="2023-08-24"),
        ui.input_date_range(
            "input_date_range", "Date range", start="2023-08-24", end="2023-08-25"
        ),
        ui.input_numeric(
            "input_numeric", "Numeric", 1, min=0, max=len(letters), step=1
        ),
        ui.input_password("input_password", "Password"),
        ui.input_radio_buttons("input_radio_buttons", "Radio buttons", choices=letters),
        ui.input_select("input_select", "Select", choices=letters),
        ui.input_selectize("input_selectize", "Selectize", choices=letters),
        ui.input_slider(
            "input_slider", "Slider", min=0, max=len(letters), step=1, value=1
        ),
        ui.input_switch("input_switch", "Switch"),
        ui.input_text("input_text", "Text"),
        ui.input_text_area("input_text_area", "Text area"),
        ui.p("(Navbars are in Cards to give some sort of bounding box.)"),
        *[
            ui.card(y)
            for y in (
                ui.navset_bar(
                    *ui_navs("Navset Bar"), title="Navset Bar", id="navset_bar"
                ),
                ui.navset_card_pill(
                    *ui_navs("Navset Card Pill"), id="navset_card_pill"
                ),
                ui.navset_card_tab(*ui_navs("Navset Card Tab"), id="navset_card_tab"),
                TagList(
                    ui.tags.label("(Navset hidden)"),
                    ui.navset_hidden(*ui_navs("Navset Hidden"), id="navset_hidden"),
                ),
                ui.navset_pill(*ui_navs("Navset Pill"), id="navset_pill"),
                ui.navset_tab(*ui_navs("Navset Tab"), id="navset_tab"),
            )
        ],
        *[
            ui.card(ui.tags.label(ui.markdown(f"{label}:")), y)
            for label, y in (
                (
                    "Data Frame (Grows w/ `input.radio_buttons()`)",
                    ui.output_data_frame("out_data_frame"),
                ),
                (
                    "Table (Grows w/ `input.radio_buttons()`)",
                    ui.output_table("out_table"),
                ),
                (
                    "Image (Changes w/ `input.radio_buttons()`)",
                    ui.output_image("out_image", height="180px"),
                ),
                ("Text Verbatim", ui.output_text_verbatim("out_text_verbatim")),
                ("Text", ui.output_text("out_text")),
                ("UI", ui.output_ui("out_ui")),
            )
        ],
        # ui.page_navbar(), # Should not be tested in a module as it is a top level element
    )


@module.server
def mod_x_server(
    input: Inputs,
    output: Outputs,
    session: Session,
):
    session_dict[session.ns] = session

    @reactive.calc
    def n():
        return int(letters.index(input.input_radio_buttons()))

    @render.image
    def out_image() -> ImgData:
        return {"src": penguin_imgs[n()]}

    @render.plot
    def out_plot():
        dt = [1, 2, 3, 4, 5]
        plt.bar(  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
            dt[: ((n() + 1) % len(dt))], dt[: ((n() + 1) % len(dt))]  # pyright: ignore
        )

    @render.table
    def out_table():
        return pandas_df.head(n() + 1)

    @render.data_frame
    def out_data_frame():
        return pandas_df.head(n() + 1)

    @render.text
    def out_text():
        return f"Output text content. `input.radio_buttons()`: `{input.input_radio_buttons()}`"

    @render.text
    def out_text_verbatim():
        return f"Output text verbatim content. `input.radio_buttons()`: `{input.input_radio_buttons()}`"

    @render.ui
    def out_ui():
        return ui.p(
            ui.markdown(
                f"Output UI content. `input.radio_buttons()`: `{input.input_radio_buttons()}`"
            )
        )

    download_button_count = 0

    @render.download(filename=lambda: f"download_button-{session.ns}.csv")
    async def download_button():
        nonlocal download_button_count
        download_button_count += 1
        yield "session,type,count\n"
        yield f"{session.ns},button,{download_button_count}\n"

    download_link_count = 0

    @render.download(filename=lambda: f"download_link-{session.ns}.csv")
    async def download_link():
        nonlocal download_link_count
        download_link_count += 1
        yield "session,type,count\n"
        yield f"{session.ns},link,{download_link_count}\n"

    def render_input_txt(
        input_key: str,
        prefix_txt: str = "",
        prefix_id: str = "",
        preprocess: Callable[[object], str] | None = None,
    ) -> None:
        @output(id=f"status_{prefix_id}{input_key}")
        @render.text
        def _():
            value = getattr(input, input_key)()
            if preprocess is not None:
                value = preprocess(value)
            return f"{prefix_txt}ui.{input_key}: `{value}`"

    for x_input_key in x_input_keys:
        render_input_txt(
            x_input_key,
            prefix_txt="x.",
            prefix_id="x_",
        )

    for input_key in input_keys:
        # Manually handle `input_file`
        if input_key == "input_file":
            continue
        render_input_txt(input_key)

    def preprocess_file(x: Any) -> str:
        if isinstance(x, list):
            # Extract the name to avoid the temp folder path in output
            return typing.cast(
                str,
                x[0].get("name"),  # pyright: ignore[reportUnknownMemberType]
            )
        else:
            return str(x)

    render_input_txt("input_file", preprocess=preprocess_file)


# =============================================================================
# App that uses module
# =============================================================================
app_ui = ui.page_fluid(
    ui.input_action_button("reset", "Reset"),
    ui.input_action_button("update_global", "Update Global"),
    ui.input_action_button("update_mod1", "Update Module 1"),
    ui.input_action_button("update_mod2", "Update Module 2"),
    ui.tags.head(
        ui.tags.style(
            """
        .popover.popover {
            --bs-popover-max-width: 500px;
        }
        """
        )
    ),
    ui.popover(
        ui.input_action_button("ignore", "Explanation of test"),
        ui.markdown(
            """
            E2E test app added has a main session and two module sessions that have inputs and outputs. It makes sure that one module does not update the other module or the main session.

            If a component does not use `resolve_id()`, the global session will be updated and the module will be unchanged.

            ### Approach
            * Reset all sessions
            * Verify all sessions have been reset
            * Update the second module x3 times
            * Verify that main session and first module have not updated
            * Verify the second module has been updated

            ### Future addtions
            * What qualifies:
                * If the input/output/update method uses `resolve_id()` or `id` (outside of `session.send_custom_input(id=)`)
                * It can be placed into a module. (e.g. not `ui.page_navbar()`)
            * Inputs
                * Add the input into the `input_keys` (or `x_input_keys`) tuple
                * Add the input into the module ui definition
                * The input `id=` should be the same as the function name. Ex: `ui.input_action_button("input_action_button", "Title")`
                * Update `expect_mod_state()` to support the new input
                * Update `expect_default_mod_state()` to reflect the reset state of the new input
            * Outputs
                * Add the output into the module ui definition
                * If it needs to be on the screen to be updated (rare), add it near the top of the definition
                * Update `expect_outputs()` to support the new output
            """
        ),
        id="explanation",
    ),
    ui.layout_column_wrap(
        mod_x_ui("", "Global"),  # "" == Root
        mod_x_ui("mod1", "Module 1"),
        mod_x_ui("mod2", "Module 2"),
        width=1 / 3,
    ),
    # ui.h3("Inputs that are not in a module:"),
    # ui.output_text_verbatim("not_modules", placeholder=True),
)


def server(input: Inputs, output: Outputs, session: Session):
    mod_x_server(session.ns)  # Root
    mod_x_server("mod1")
    mod_x_server("mod2")

    # Check if in interactive mode
    # https://stackoverflow.com/a/64523765/591574
    if hasattr(sys, "ps1"):
        # Open the explanation popover
        ui.update_popover("explanation", show=True)

        # On button clicks, hide the explanation popover
        @reactive.effect(suspended=True)
        def _():
            input.reset()
            input.update_global()
            input.update_mod1()
            input.update_mod2()
            ui.update_popover("explanation", show=False)

    # Master function to update a module's features
    def update_session(
        session: Session,
        *,
        count: int = 1,
    ):
        letter = letters[count % len(letters)]
        on_off = count % 2 == 1
        with session_context(session):
            ui.update_accordion("accordion", show=letter)

            ui.update_checkbox("input_checkbox", value=on_off)
            checkbox_group_letters = letters.copy()
            checkbox_group_letters.remove(letter)
            ui.update_checkbox_group(
                "input_checkbox_group", selected=checkbox_group_letters
            )
            date_start = datetime.date(2023, 8, 24) + datetime.timedelta(days=count)
            ui.update_date("input_date", value=date_start)
            ui.update_date_range(
                "input_date_range",
                start=date_start + datetime.timedelta(days=1),
                end=date_start + datetime.timedelta(days=3),
            )
            ui.update_numeric("input_numeric", value=count)
            ui.update_text("input_password", value=f"password{count}")
            ui.update_radio_buttons("input_radio_buttons", selected=letter)
            ui.update_select("input_select", selected=letter)
            ui.update_selectize("input_selectize", selected=letter)
            ui.update_slider("input_slider", value=count)
            ui.update_text("input_text", value=f"text{count}")
            ui.update_text_area("input_text_area", value=f"text_area{count}")

            # Toggle elements
            if count == 0:
                # Open sidebar
                # https://github.com/posit-dev/py-shiny/issues/716
                # ui.update_sidebar("sidebar", value=True)
                if session.input.sidebar() is False:
                    ui.update_sidebar("sidebar", show=True)

                # Turn off switch
                ui.update_switch("input_switch", value=False)

                # Hide popover
                # https://github.com/posit-dev/py-shiny/issues/717
                # ui.update_popover("popover", show=False)
                if session.input.popover() is True:
                    ui.update_popover("popover", show=False)
                # Hide tooltip
                # https://github.com/posit-dev/py-shiny/issues/717
                if session.input.tooltip() is True:
                    ui.update_tooltip("tooltip", show=False)
            else:
                if session.input.sidebar() == on_off:
                    ui.update_sidebar("sidebar", show=not on_off)
                if session.input.input_switch() != on_off:
                    ui.update_switch("input_switch", value=on_off)
                if session.input.popover() != on_off:
                    ui.update_popover("popover", show=on_off)
                if session.input.tooltip() != on_off:
                    ui.update_tooltip("tooltip", show=on_off)

            ui.update_navs("navset_bar", selected=letter)
            ui.update_navs("navset_card_pill", selected=letter)
            ui.update_navs("navset_card_tab", selected=letter)
            ui.update_navs("navset_hidden", selected=letter)
            ui.update_navs("navset_pill", selected=letter)
            ui.update_navs("navset_tab", selected=letter)

    module_keys = (
        ("update_global", ""),
        ("update_mod1", "mod1"),
        ("update_mod2", "mod2"),
    )

    offsets: dict[str, int] = {}
    for _input_key, session_key in module_keys:
        offsets[session_key] = 0

    @reactive.effect
    def _():
        # trigger reactivity
        input.reset()
        with reactive.isolate():
            for input_key, session_key in module_keys:
                offsets[session_key] = input[input_key]()
                update_session(session_dict[session_key], count=0)

    def update_session_effect(*, input_key: str, session_key: str) -> None:
        @reactive.effect
        @reactive.event(input[input_key])
        def _():
            update_session(
                session_dict[session_key],
                count=input[input_key]() - offsets[session_key],
            )

    for input_key, session_key in module_keys:
        update_session_effect(input_key=input_key, session_key=session_key)

    # # ## Debug

    # @reactive.effect
    # def _():
    #     print("here")
    #     reactive.invalidate_later(1)
    #     print(input._ns)
    #     print(
    #         [
    #             key
    #             for key in input._map.keys()
    #             if not key.startswith(".") and not key.startswith("mod")
    #         ]
    #     )


app = App(app_ui, server)



from __future__ import annotations

import pkgutil

# pyright: reportMissingTypeStubs = false
import palmerpenguins
import polars as pl
from narwhals.stable.v1.typing import IntoDataFrame

from shiny import App, Inputs, Outputs, Session, module, reactive, render, ui

# Load the dataset


pd_penguins = palmerpenguins.load_penguins_raw().iloc[0:5, 0:6]
pl_penguins = pl.read_csv(
    pkgutil.get_data(  # pyright: ignore[reportArgumentType]
        "palmerpenguins",
        "data/penguins-raw.csv",
    )
)[["studyName", "Sample Number", "Species", "Region", "Island", "Stage"]].head(5)


# import great_tables as gt
# from great_tables_test_utils import gt_styles

# df_gt = (
#     gt.GT(df)
#     .tab_style(
#         [
#             gt.style.fill(color="lightblue"),
#         ],
#         gt.loc.body(),
#     )
#     .tab_style(
#         [
#             gt.style.fill(color="purple"),
#             gt.style.borders(color="green", style="dashed"),
#         ],
#         gt.loc.body("Species", [1, 2]),
#     )
#     .tab_style(
#         gt.style.fill(color="yellow"),
#         gt.loc.body("Region", [2]),
#     )
#     .tab_style(
#         gt.style.fill(color="red"),
#         gt.loc.body("Island"),
#     )
# )
# df_styles = gt_styles(df_gt)

df_styles: list[render.StyleInfo] = [
    {
        "style": {"color": "darkorange", "font-weight": "bold"},
    },
    {
        "rows": None,
        "cols": None,
        "style": {"background-color": "lightblue"},
    },
    {
        "rows": [1, 2],
        "cols": "Species",
        "style": {
            "background-color": "purple",
            "border-color": "green",
            "border-style": "dashed",
        },
    },
    {
        "location": "body",
        "rows": 2,
        "cols": ["Region"],
        "style": {"background-color": "yellow"},
    },
    {
        "location": "body",
        "rows": None,
        "cols": [4],  # "Island",
        "style": {"background-color": "red"},
    },
    {
        "location": "body",
        "rows": [False, True, False, False, False],
        "cols": [False, False, False, False, True, True],  # "Stage",
        "style": {"background-color": "green"},
    },
]


@module.ui
def mod_ui():
    return ui.TagList(
        ui.markdown(
            "**Instructions**: Edit the cells 3 times. Watch the styles change in the first card."
        ),
        ui.card(
            ui.card_header("Styles Function:"),
            ui.output_data_frame("fn_styles"),
        ),
        ui.card(
            ui.card_header("Styles List:"),
            ui.output_data_frame("list_styles"),
        ),
    )


@module.server
def mod_server(
    input: Inputs,
    output: Outputs,
    session: Session,
    data: IntoDataFrame,
):
    @render.data_frame
    def fn_styles():

        def df_styles_fn(
            data: IntoDataFrame,
        ) -> list[render.StyleInfo]:

            def style_is_everywhere(style_info: render.StyleInfo):
                return (style_info.get("rows", None) is None) and (
                    style_info.get("cols", None) is None
                )

            everywhere_styles = [s for s in df_styles if style_is_everywhere(s)]

            with reactive.isolate():
                patch_size = len(fn_styles._cell_patch_map().keys()) + 1
            if patch_size > len(df_styles) - len(everywhere_styles):
                patch_size = 1

            ret: list[render.StyleInfo] = []
            for style_info in df_styles:
                style_val = style_info.get("style", None)
                if style_val is None:
                    continue
                if style_is_everywhere(style_info):
                    continue
                ret.append(style_info)
                if len(ret) >= patch_size:
                    break
            return ret

        # NOTE - Styles in GT are greedy!
        # Q: Cell editing should not be allowed as styles would be applied given the original data
        #   * Once GT becomes not greedy, we can unlock this.
        # Styles can be subsetted by looking for the row value
        # return df
        return render.DataTable(
            data,
            selection_mode=("rows"),
            editable=True,
            # filters=True,
            styles=df_styles_fn,
        )
        # # Ideal
        # return render.DataGrid(
        #     df_gt,
        #     selection_mode=("rows"),
        #     editable=True,
        # )

    @render.data_frame
    def list_styles():
        return render.DataTable(
            data,
            selection_mode=("rows"),
            editable=True,
            # filters=True,
            styles=df_styles,
        )


app_ui = ui.page_fillable(
    ui.navset_underline(
        ui.nav_panel("pandas", mod_ui("pandas")),
        ui.nav_panel("polars", mod_ui("polars")),
        id="tab",
    )
)


def server(input: Inputs):
    mod_server("pandas", pd_penguins)
    mod_server("polars", pl_penguins)


app = App(app_ui, server, debug=False)



# pyright: reportUnknownMemberType=false

import re
import time
from typing import Any, Callable

import pytest
from conftest import create_example_fixture
from examples.example_apps import reruns, reruns_delay
from playwright.sync_api import Locator, Page, expect

from shiny.playwright import controller
from shiny.playwright.expect import expect_to_change
from shiny.run import ShinyAppProc

data_frame_app = create_example_fixture("dataframe")


@pytest.fixture
def grid(page: Page) -> Locator:
    return page.locator("#grid")


@pytest.fixture
def grid_container(page: Page, grid: Locator) -> Locator:
    return grid.locator("> div > div.shiny-data-grid")


@pytest.fixture
def summary(page: Page, grid: Locator) -> Locator:
    return grid.locator("div.shiny-data-grid-summary")


@pytest.fixture
def scroll_to_end(page: Page, grid_container: Locator) -> Callable[[], None]:
    def do():
        grid_container.locator("tbody tr:first-child td:first-child").click()
        page.keyboard.press("End")

        # Starting some time around January 19, 2024, Firefox doesn't scroll to the end
        # of the table when the End button is pressed. Repros on real Firefox 121.0.1
        # (64-bit) on macOS. This is a workaround to get the test suite to pass.
        if page.context.browser and page.context.browser.browser_type.name == "firefox":
            time.sleep(0.1)
            page.keyboard.press("End")
            time.sleep(0.1)
            page.keyboard.press("End")

    return do


@pytest.mark.flaky(reruns=reruns, delay=reruns_delay)
def test_grid_mode(
    page: Page, data_frame_app: ShinyAppProc, grid: Locator, grid_container: Locator
):
    page.goto(data_frame_app.url)

    # Ensure the output was found by Shiny
    expect(grid).to_have_class(re.compile(r"\bshiny-bound-output\b"))

    # Expect the data frame to be in grid mode by default
    expect(grid_container).not_to_have_class(re.compile(r"\bshiny-data-grid-table\b"))
    expect(grid_container).to_have_class(re.compile(r"\bshiny-data-grid-grid\b"))


@pytest.mark.flaky(reruns=reruns, delay=reruns_delay)
def test_summary_navigation(
    page: Page,
    data_frame_app: ShinyAppProc,
    grid_container: Locator,
    summary: Locator,
    scroll_to_end: Callable[[], None],
):
    page.goto(data_frame_app.url)

    # Check that summary responds to navigation
    expect(summary).to_have_text(re.compile("^Viewing rows 1 through \\d+ of 20$"))
    # Put focus in the table and hit End keystroke
    grid_container.locator("tbody tr:first-child td:first-child").click()
    with expect_to_change(lambda: summary.inner_text()):
        scroll_to_end()
    # Ensure that summary updated
    expect(summary).to_have_text(re.compile("^Viewing rows \\d+ through 20 of 20$"))


@pytest.mark.flaky(reruns=reruns, delay=reruns_delay)
def test_full_width(page: Page, data_frame_app: ShinyAppProc, grid_container: Locator):
    page.goto(data_frame_app.url)

    def get_width() -> float:
        rect = grid_container.bounding_box()
        return rect.get("width", 0) if rect else 0

    width1 = get_width()
    # Switch to narrow mode
    with expect_to_change(get_width):
        controller.InputSwitch(page, "fullwidth")._toggle()
    width2 = get_width()

    assert width2 < width1

    # Switch back to full width
    controller.InputSwitch(page, "fullwidth")._toggle()


@pytest.mark.flaky(reruns=reruns, delay=reruns_delay)
def test_table_switch(
    page: Page,
    data_frame_app: ShinyAppProc,
    grid: Locator,
    grid_container: Locator,
    summary: Locator,
    scroll_to_end: Callable[[], None],
):
    page.goto(data_frame_app.url)
    select_dataset = controller.InputSelect(page, "dataset")

    scroll_to_end()

    # Switch to table
    controller.InputSwitch(page, "gridstyle")._toggle()
    expect(grid_container).not_to_have_class(re.compile(r"\bshiny-data-grid-grid\b"))
    expect(grid_container).to_have_class(re.compile(r"\bshiny-data-grid-table\b"))

    # Switching modes resets scroll
    expect(summary).to_have_text(re.compile("^Viewing rows 1 through \\d+ of 20$"))

    scroll_to_end()
    expect(summary).to_have_text(re.compile("^Viewing rows \\d+ through 20 of 20$"))

    # Switch datasets to much longer one
    select_dataset.set("diamonds")
    select_dataset.expect.to_have_value("diamonds")
    expect(summary).to_have_text(
        re.compile("^Viewing rows 1 through \\d+ of 53940$"), timeout=10000
    )


@pytest.mark.flaky(reruns=reruns, delay=reruns_delay)
def test_sort(
    page: Page,
    data_frame_app: ShinyAppProc,
    grid_container: Locator,
):
    page.goto(data_frame_app.url)
    select_dataset = controller.InputSelect(page, "dataset")
    select_dataset.set("diamonds")
    select_dataset.expect.to_have_value("diamonds")

    # Table cell locators
    header_clarity = grid_container.locator("tr:first-child th:nth-child(4)")
    first_cell_clarity = grid_container.locator("tr:first-child td:nth-child(4)")

    # Test that the table contents have updated
    # This may timeout unless a larger timeout is given
    expect(header_clarity).not_to_have_text(
        "num2",
        timeout=15 * 1000,  # Larger timeout for CI
    )
    expect(first_cell_clarity).not_to_have_text("4")

    # Test sorting
    expect(first_cell_clarity).to_have_text("SI2")
    header_clarity.click()
    expect(first_cell_clarity).to_have_text("I1")
    header_clarity.click()
    expect(first_cell_clarity).to_have_text("VVS2")

    # Test multi-sorting
    depth_clarity = grid_container.locator("tr:first-child th:nth-child(5)")
    first_cell_depth = grid_container.locator("tr:first-child td:nth-child(5)")
    depth_clarity.click(modifiers=["Shift"])
    expect(first_cell_depth).to_have_text("67.6")


@pytest.mark.flaky(reruns=reruns, delay=reruns_delay)
def test_multi_selection(
    page: Page,
    data_frame_app: ShinyAppProc,
    grid_container: Locator,
    snapshot: Any,
):
    page.goto(data_frame_app.url)
    controller.InputSelect(page, "selection_mode").set("rows")
    first_cell = grid_container.locator("tbody tr:first-child td:first-child")

    def detail_text():
        return page.locator("#detail").inner_text()

    expect(first_cell).to_have_text("1")
    with expect_to_change(detail_text):
        first_cell.click()

    assert detail_text() == snapshot

    kb = page.keyboard
    with expect_to_change(lambda: grid_container.locator(":focus").inner_text()):
        kb.press("ArrowDown")
    with expect_to_change(lambda: grid_container.locator(":focus").inner_text()):
        kb.press("ArrowDown")
    with expect_to_change(detail_text):
        kb.press("Space")
    with expect_to_change(lambda: grid_container.locator(":focus").inner_text()):
        kb.press("ArrowDown")
    with expect_to_change(detail_text):
        kb.press("Enter")
    assert detail_text() == snapshot


@pytest.mark.flaky(reruns=reruns, delay=reruns_delay)
def test_single_selection(
    page: Page,
    data_frame_app: ShinyAppProc,
    grid_container: Locator,
    snapshot: Any,
):
    page.goto(data_frame_app.url)
    controller.InputSelect(page, "selection_mode").set("row")
    first_cell = grid_container.locator("tbody tr:first-child td:first-child")

    def detail_text():
        return page.locator("#detail").inner_text()

    expect(first_cell).to_have_text("1")
    with expect_to_change(detail_text):
        first_cell.click()

    assert detail_text() == snapshot

    kb = page.keyboard
    with expect_to_change(detail_text):
        kb.press("ArrowDown")
    with expect_to_change(detail_text):
        kb.press("ArrowDown")
    with expect_to_change(detail_text):
        kb.press("Space")
    with expect_to_change(lambda: grid_container.locator(":focus").inner_text()):
        kb.press("ArrowDown")
    with expect_to_change(detail_text):
        kb.press("Enter")
    assert detail_text() == snapshot


def test_filter_grid(
    page: Page,
    data_frame_app: ShinyAppProc,
    grid: Locator,
    summary: Locator,
    snapshot: Any,
):
    page.goto(data_frame_app.url)
    _filter_test_impl(page, data_frame_app, grid, summary, snapshot)


def test_filter_table(
    page: Page,
    data_frame_app: ShinyAppProc,
    grid: Locator,
    grid_container: Locator,
    summary: Locator,
    snapshot: Any,
):
    page.goto(data_frame_app.url)

    controller.InputSwitch(page, "gridstyle")._toggle()
    expect(grid_container).not_to_have_class(re.compile(r"\bshiny-data-grid-grid\b"))
    expect(grid_container).to_have_class(re.compile(r"\bshiny-data-grid-table\b"))

    _filter_test_impl(page, data_frame_app, grid, summary, snapshot)


def _filter_test_impl(
    page: Page,
    data_frame_app: ShinyAppProc,
    grid: Locator,
    summary: Locator,
    snapshot: Any,
):
    controller.InputSelect(page, "selection_mode").set("none")

    filters = grid.locator("tr.filters")

    filter_subidir_min = filters.locator("> th:nth-child(1) > div > input:nth-child(1)")
    filter_subidir_max = filters.locator("> th:nth-child(1) > div > input:nth-child(2)")
    filter_attnr = filters.locator("> th:nth-child(2) > input")
    filter_num1_min = filters.locator("> th:nth-child(3) > div > input:nth-child(1)")
    filter_num1_max = filters.locator("> th:nth-child(3) > div > input:nth-child(2)")

    expect(filter_subidir_min).to_be_visible()
    expect(filter_subidir_max).to_be_visible()
    expect(filter_attnr).to_be_visible()
    expect(filter_num1_min).to_be_visible()
    expect(filter_num1_max).to_be_visible()

    expect(summary).to_be_visible()
    expect(summary).to_have_text(re.compile(" of 20$"))

    # Placeholder text only appears when filter is focused
    expect(page.get_by_placeholder("Min (1)", exact=True)).not_to_be_attached()
    expect(page.get_by_placeholder("Max (20)", exact=True)).not_to_be_attached()
    filter_subidir_min.focus()
    expect(page.get_by_placeholder("Min (1)", exact=True)).to_be_attached()
    expect(page.get_by_placeholder("Max (20)", exact=True)).to_be_attached()
    filter_subidir_min.blur()
    expect(page.get_by_placeholder("Min (1)", exact=True)).not_to_be_attached()
    expect(page.get_by_placeholder("Max (20)", exact=True)).not_to_be_attached()

    # Make sure that filtering input results in correct number of rows

    # Test only min
    filter_subidir_min.fill("5")
    expect(summary).to_have_text(re.compile(" of 16$"))
    # Test min and max
    filter_subidir_max.fill("14")
    expect(summary).to_have_text(re.compile(" of 10$"))

    # When filtering results in all rows being shown, the summary should not be visible
    filter_subidir_max.fill("11")
    expect(summary).not_to_be_attached()

    # Test only max
    filter_subidir_min.fill("")
    expect(summary).to_have_text(re.compile(" of 11"))

    filter_subidir_min.clear()
    filter_subidir_max.clear()

    # Try substring search
    filter_attnr.fill("oc")
    expect(summary).to_have_text(re.compile(" of 10"))
    filter_num1_min.focus()
    # Ensure other columns' filter placeholders show faceted results
    expect(page.get_by_placeholder("Min (5)", exact=True)).to_be_attached()
    expect(page.get_by_placeholder("Max (8)", exact=True)).to_be_attached()

    # Filter down to zero matching rows
    filter_attnr.fill("q")
    # Summary should be gone
    expect(summary).not_to_be_attached()
    filter_num1_min.focus()
    # Placeholders should not have values
    expect(page.get_by_placeholder("Min", exact=True)).to_be_attached()
    expect(page.get_by_placeholder("Max", exact=True)).to_be_attached()

    filter_attnr.clear()

    # Apply multiple filters, make sure we get the correct results
    filter_subidir_max.fill("8")
    # We had a bug before where typing in a decimal point would cause
    # the cursor to jump to the front. Make sure that doesn't happen.
    filter_num1_min.focus()
    page.keyboard.press("3")
    time.sleep(0.2)
    page.keyboard.press(".")
    time.sleep(0.2)
    page.keyboard.press("9")
    expect(grid.locator("tbody tr")).to_have_count(5)

    # Ensure changing dataset resets filters
    select_dataset = controller.InputSelect(page, "dataset")
    select_dataset.set("attention")
    select_dataset.expect.to_have_value("attention")
    expect(page.get_by_text("Unnamed: 0")).to_be_attached()
    select_dataset.set("anagrams")
    select_dataset.expect.to_have_value("anagrams")
    expect(summary).to_have_text(re.compile(" of 20"))


def test_filter_disable(page: Page, data_frame_app: ShinyAppProc):
    page.goto(data_frame_app.url)

    expect(page.locator("tr.filters")).to_be_attached()
    controller.InputSwitch(page, "filters")._toggle()
    expect(page.locator("tr.filters")).not_to_be_attached()



from datetime import datetime

import htmltools
import pandas as pd
import polars as pl

from shiny.express import render, ui

DATA = {
    "num": [1, 2],
    "chr": ["a", "b"],
    "cat": ["a", "a"],
    "dt": [datetime(2000, 1, 2)] * 2,
    "html": [htmltools.em("emphasized")] * 2,
    "html_str": [htmltools.HTML("<strong>bolded</strong>")] * 2,
    # "struct": [{"x": 1}, {"x": 2}],
}

ui.h2("data_frame column types")

with ui.card():
    ui.card_header("Pandas")

    @render.data_frame
    def pd_table():
        return render.DataGrid(pd.DataFrame(DATA), filters=True)


with ui.card():
    ui.card_header("Polars")

    @render.data_frame
    def pl_table():
        return render.DataGrid(pl.DataFrame(DATA), filters=True)



from shiny import App, Inputs, Outputs, Session, ui

app_ui = ui.page_fluid(
    ui.panel_title("A basic absolute panel example", "Demo"),
    ui.panel_absolute(
        ui.panel_well(
            "Drag me around!", ui.input_slider("n", "N", min=0, max=100, value=20)
        ),
        draggable=True,
        width="300px",
        right="50px",
        top="25%",
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    pass


app = App(app_ui, server)



# pyright: reportUnknownArgumentType=false

from __future__ import annotations

__all__ = ("brushed_points", "near_points")


from typing import TYPE_CHECKING, Literal, Optional, Union, cast

from ._typing_extensions import TypedDict
from .types import BrushInfo, CoordInfo, CoordXY

if TYPE_CHECKING:
    import numpy.typing as npt
    import pandas as pd

DataFrameColumn = Union[
    "pd.Series[int]",
    "pd.Series[float]",
    "pd.Series[str]",
    "pd.Categorical",
    "pd.DatetimeIndex",
]


class SeriesFloatXY(TypedDict):
    x: pd.Series[float]
    y: pd.Series[float]


def brushed_points(
    df: pd.DataFrame,
    brush: BrushInfo | None,
    xvar: Optional[str] = None,
    yvar: Optional[str] = None,
    panelvar1: Optional[str] = None,
    panelvar2: Optional[str] = None,
    *,
    all_rows: bool = False,
) -> pd.DataFrame:
    """Find rows of data selected on an interactive plot.

    This function is used with interactive plots. It returns the rows of a data frame
    which are under a brush.

    It currently supports plots created by matplotlib, seaborn, and plotnine. If
    plotnine is used, it can usually automatically infer the x and y variables, along
    with variables used for facets.

    Parameters
    ----------
    df
        A pandas DataFrame from which to select rows.
    brush
        The data from a brush, like `input.myplot_brush()`.
    xvar
        The name of the column in `df` that contains the x values. (Note that when using
        plotnine, `xvar`, `yvar`, `panelvar1`, and `panelvar2` can usually be
        automatically inferred from the brush data.)
    yvar
        The name of the column in `df` that contains the y values.
    panelvar1
        The name of the column in `df` that contains the first variable used for
        subpanels (if subpanels are used).
    panelvar2
        The name of the column in `df` that contains the second variable used for
        subpanels.
    all_rows
        If `False` (the default), return a data frame containing only the rows that are
        selected. If `True`, then all rows from the data frame will be returned, along
        with an additional column named `selected_`, which indicates whether or not each
        row was selected.

    Returns
    -------
    :
        A pandas DataFrame containing the rows selected by the brush. If `all_rows` is
        `True`, then all rows from the original data will be returned, along with an
        additional column named `selected_`, which indicates whether or not each row was
        selected.
    """
    import pandas as pd

    new_df = df.copy()

    if brush is None:
        if all_rows:
            new_df["selected_"] = False
        else:
            new_df = new_df.loc[[]]

        return new_df

    if "xmin" not in brush:
        raise ValueError(
            "brushed_points requires a brush object with xmin, xmax, ymin, and ymax."
        )

    # Which direction(s) the brush is selecting over. Direction can be 'x', 'y',
    # or 'xy'.
    use_x = "x" in brush["direction"]
    use_y = "y" in brush["direction"]

    # Filter out x and y values
    keep_rows: pd.Series[bool] = pd.Series(
        True,
        index=new_df.index,  # pyright: ignore[reportUnknownMemberType]
    )
    if use_x:
        if xvar is None and "x" in brush["mapping"]:
            xvar = brush["mapping"]["x"]
        if xvar is None:
            raise ValueError(
                "brushed_points: not able to automatically infer `xvar` from brush. You must supply `xvar` to brushed_points()"
            )
        if xvar not in new_df:
            raise ValueError(f"brushed_points: `xvar` ({xvar}) not in dataframe")
        keep_rows &= within_brush(new_df[xvar], brush, "x")

    if use_y:
        if yvar is None and "y" in brush["mapping"]:
            yvar = brush["mapping"]["y"]
        if yvar is None:
            raise ValueError(
                "brushed_points: not able to automatically infer `yvar` from brush. You must supply `yvar` to brushed_points()"
            )
        if yvar not in new_df:
            raise ValueError(f"brushed_points: `yvar` ({yvar}) not in dataframe")
        keep_rows &= within_brush(new_df[yvar], brush, "y")

    # Find which rows are matches for the panel vars (if present)
    if panelvar1 is None and "panelvar1" in brush["mapping"]:
        panelvar1 = brush["mapping"]["panelvar1"]
        if panelvar1 not in new_df:
            raise ValueError(
                f"brushed_points: `panelvar1` ({panelvar1}) not in dataframe"
            )
        keep_rows &= new_df[panelvar1] == brush["panelvar1"]  # pyright: ignore

    if panelvar2 is None and "panelvar2" in brush["mapping"]:
        panelvar2 = brush["mapping"]["panelvar2"]
        if panelvar2 not in new_df:
            raise ValueError(
                f"brushed_points: `panelvar2` ({panelvar2}) not in dataframe"
            )
        keep_rows &= new_df[panelvar2] == brush["panelvar2"]  # pyright: ignore

    if all_rows:
        new_df["selected_"] = False
        new_df.loc[keep_rows, "selected_"] = True
    else:
        new_df = new_df.loc[keep_rows]

    return new_df


def near_points(
    df: pd.DataFrame,
    coordinfo: CoordInfo | None,
    xvar: Optional[str] = None,
    yvar: Optional[str] = None,
    panelvar1: Optional[str] = None,
    panelvar2: Optional[str] = None,
    *,
    threshold: float = 5,
    max_points: Optional[int] = None,
    add_dist: bool = False,
    all_rows: bool = False,
) -> pd.DataFrame:
    """Find rows of data selected on an interactive plot.

    This function is used with interactive plots. It returns the rows of a data frame
    which are under a brush.

    It currently supports plots created by matplotlib, seaborn, and plotnine. If
    plotnine is used, it can usually automatically infer the x and y variables, along
    with variables used for facets.

    Parameters
    ----------
    df
        A pandas DataFrame from which to select rows.
    coordinfo
        The data from a click/dblclick/hover event, like `input.myplot_click()`.
    xvar
        The name of the column in `df` that contains the x values. (Note that when using
        plotnine, `xvar`, `yvar`, `panelvar1`, and `panelvar2` can usually be
        automatically inferred from the brush data.)
    yvar
        The name of the column in `df` that contains the y values.
    panelvar1
        The name of the column in `df` that contains the first variable used for
        subpanels (if subpanels are used).
    panelvar2
        The name of the column in `df` that contains the second variable used for
        subpanels.
    threshold
        A maximum distance (in pixels) to the pointer location. Rows in the data frame
        will be selected if the distance to the pointer is less than `threshold`.
    max_points
        Maximum number of rows to return. If `None` (the default), will return all rows
        within the threshold distance.
    add_dist
        If `True`, add a column named `dist_` that contains the distance from the
        coordinate to the point, in pixels. When no pointer event has yet occurred, the
        value of `dist_` will be `numpy.NaN`.
    all_rows
        If `False` (the default), return a data frame containing only the rows that are
        selected. If `True`, then all rows from the data frame will be returned, along
        with an additional column named `selected_`, which indicates whether or not each
        row was selected.

    Returns
    -------
    :
        A pandas DataFrame containing the rows selected by the brush. If `all_rows` is
        `True`, then all rows from the original data will be returned, along with an
        additional column named `selected_`, which indicates whether or not each row was
        selected.
    """
    import numpy as np

    new_df = df.copy()

    # For no current coordinfo
    if coordinfo is None:
        if add_dist:
            new_df["dist"] = np.nan

        if all_rows:
            new_df["selected_"] = False
        else:
            new_df = new_df.loc[[]]

        return new_df

    # Try to extract vars from coordinfo object
    coordinfo_mapping = coordinfo["mapping"]
    if xvar is None and "x" in coordinfo_mapping:
        xvar = coordinfo_mapping["x"]
    if yvar is None and "y" in coordinfo_mapping:
        yvar = coordinfo_mapping["y"]

    if xvar is None:
        if xvar is None and "x" in coordinfo["mapping"]:
            xvar = coordinfo["mapping"]["x"]
        raise ValueError(
            "near_points: not able to automatically infer `xvar` from coordinfo. You must supply `xvar` to near_points()"
        )
    if yvar is None:
        if yvar is None and "y" in coordinfo["mapping"]:
            yvar = coordinfo["mapping"]["y"]
        raise ValueError(
            "near_points: not able to automatically infer `yvar` from coordinfo. You must supply `yvar` to near_points()"
        )

    if xvar not in new_df.columns:
        raise ValueError(f"near_points: `xvar` ('{xvar}')  not in names of input.")
    if yvar not in new_df.columns:
        raise ValueError(f"near_points: `yvar` ('{yvar}')  not in names of input.")

    x: pd.Series[float] = to_float(new_df[xvar])
    y: pd.Series[float] = to_float(new_df[yvar])

    # Get the coordinates of the point (in img pixel coordinates)
    point_img: CoordXY = coordinfo["coords_img"]

    # Get coordinates of data points (in img pixel coordinates)
    data_img: SeriesFloatXY = scale_coords(x, y, coordinfo)

    # Get x/y distances (in css coordinates)
    dist_css: SeriesFloatXY = {
        "x": (data_img["x"] - point_img["x"]) / coordinfo["img_css_ratio"]["x"],
        "y": (data_img["y"] - point_img["y"]) / coordinfo["img_css_ratio"]["y"],
    }

    # Distances of data points to the target point, in css pixels.
    dists: pd.Series[float] = (dist_css["x"] ** 2 + dist_css["y"] ** 2) ** 0.5

    if add_dist:
        new_df["dist"] = dists

    keep_rows = dists <= threshold

    # Find which rows are matches for the panel vars (if present)
    if panelvar1 is None and "panelvar1" in coordinfo["mapping"]:
        panelvar1 = coordinfo["mapping"]["panelvar1"]
        if panelvar1 not in new_df:
            raise ValueError(f"near_points: `panelvar1` ({panelvar1}) not in dataframe")
        keep_rows &= new_df[panelvar1] == coordinfo["panelvar1"]  # pyright: ignore

    if panelvar2 is None and "panelvar2" in coordinfo["mapping"]:
        panelvar2 = coordinfo["mapping"]["panelvar2"]
        if panelvar2 not in new_df:
            raise ValueError(f"near_points: `panelvar2` ({panelvar2}) not in dataframe")
        keep_rows &= new_df[panelvar2] == coordinfo["panelvar2"]  # pyright: ignore

    # Track the row indices to keep (note this is the row position, 0, 1, 2, not the
    # pandas index column, which can have arbitrary values).
    keep_idx = np.where(keep_rows)[0]

    # Order by distance
    dists = dists.iloc[keep_idx]
    keep_idx: npt.NDArray[np.intp] = keep_idx[dists.argsort()]

    # Keep max number of rows
    if max_points is not None and len(keep_idx) > max_points:
        keep_idx = keep_idx[:max_points]

    if all_rows:
        # Add selected_ column if needed
        new_df["selected_"] = False
        new_df.iloc[  # pyright: ignore[reportArgumentType]
            keep_idx,
            new_df.columns.get_loc(  # pyright: ignore[reportUnknownMemberType]
                "selected_"
            ),
        ] = True
    else:
        new_df = new_df.iloc[keep_idx]

    return new_df


# ===============================================================================
# Helper functions
# ===============================================================================
# Helper to determine if data values are within the limits of
# an input brush.
def within_brush(
    vals: DataFrameColumn,
    brush: BrushInfo,
    var: Literal["x", "y"] = "x",
) -> pd.Series[bool]:
    vals = to_float(vals)
    return (vals >= brush[var + "min"]) & (vals <= brush[var + "max"])


def to_float(x: DataFrameColumn) -> pd.Series[float]:
    import pandas as pd
    import pandas.api.types as ptypes

    """Convert int/float/str/categorical Series to a float Series.

    If the input is a int or float Series, this returns a copy. Otherwise, it returns a
    new Series object.
    """
    if ptypes.is_numeric_dtype(x):  # pyright: ignore[reportUnknownMemberType]
        return cast("pd.Series[float]", x)
    elif isinstance(x, ptypes.CategoricalDtype):
        return cast("pd.Series[float]", x.cat.codes + 1)  # pyright: ignore
    elif ptypes.is_string_dtype(x):  # pyright: ignore[reportUnknownMemberType]
        return cast(
            "pd.Series[float]", x.astype("category").cat.codes + 1  # pyright: ignore
        )
    elif ptypes.is_datetime64_any_dtype(x):  # pyright: ignore[reportUnknownMemberType]
        # We need to convert the pandas datetimes, which are in nanoseconds since epoch,
        # to matplotlib datetimes, which are in days since epoch..
        return (
            cast(
                "pd.Series[int]",
                pd.Series(x.astype("int64")),  # pyright: ignore
            )
            / (24 * 60 * 60)
            / 1e9
        )

    raise ValueError("to_float: unsupported dtype for x")


# ===============================================================================
# Scaling functions
# ===============================================================================
# These functions have direct analogs in Javascript code, except these are
# vectorized for x and y.


# Map a value x from a domain to a range. If clip is true, clip it to the
# range.
def map_linear(
    x: pd.Series[float],
    domain_min: float,
    domain_max: float,
    range_min: float,
    range_max: float,
    clip: bool = True,
) -> pd.Series[float]:
    factor = (range_max - range_min) / (domain_max - domain_min)
    val: pd.Series[float] = x - domain_min
    newval: pd.Series[float] = (val * factor) + range_min

    if clip:
        maxval = max(range_max, range_min)
        minval = min(range_max, range_min)
        newval[newval > maxval] = maxval
        newval[newval < minval] = minval

    return newval


# Scale val from domain to range. If logbase is present, use log scaling.
def scale_1d(
    val: pd.Series[float],
    domain_min: float,
    domain_max: float,
    range_min: float,
    range_max: float,
    logbase: Optional[float] = None,
    clip: bool = True,
) -> pd.Series[float]:
    import numpy as np

    if logbase is not None:
        val = np.log(val) / np.log(logbase)

    return map_linear(val, domain_min, domain_max, range_min, range_max, clip)


# Scale x and y coordinates from domain to range, using information in scaleinfo.
# scaleinfo must contain items $domain, $range, and $log. The scaleinfo object
# corresponds to one element from the coordmap object generated by getPrevPlotCoordmap
# or getGgplotCoordmap; it is the scaling information for one panel in a plot.
def scale_coords(
    x: pd.Series[float],
    y: pd.Series[float],
    coordinfo: CoordInfo,
) -> SeriesFloatXY:
    domain = coordinfo["domain"]
    range = coordinfo["range"]
    log = coordinfo["log"]

    return {
        "x": scale_1d(
            x, domain["left"], domain["right"], range["left"], range["right"], log["x"]
        ),
        "y": scale_1d(
            y, domain["bottom"], domain["top"], range["bottom"], range["top"], log["y"]
        ),
    }



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_output_transformer(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    controller.OutputTextVerbatim(page, "t1").expect_value("t1; no call; sync")
    controller.OutputTextVerbatim(page, "t2").expect_value("t2; no call; async")
    controller.OutputTextVerbatim(page, "t3").expect_value("t3; call; sync")
    controller.OutputTextVerbatim(page, "t4").expect_value("t4; call; async")
    controller.OutputTextVerbatim(page, "t5").expect_value(
        "t5; call; sync; w/ extra_txt"
    )
    controller.OutputTextVerbatim(page, "t6").expect_value(
        "t6; call; async; w/ extra_txt"
    )



import pkgutil

import palmerpenguins  # pyright: ignore[reportMissingTypeStubs]
import polars as pl
from narwhals.stable.v1.typing import IntoDataFrame

from shiny import App, Inputs, Outputs, Session, module, reactive, render, ui

pd_penguins = palmerpenguins.load_penguins_raw()
pl_penguins = pl.read_csv(
    pkgutil.get_data(  # pyright: ignore[reportArgumentType]
        "palmerpenguins", "data/penguins-raw.csv"
    )
)


def make_ui(title: str):
    return ui.card(
        title,
        ui.layout_columns(
            ui.input_select(
                "selection_mode",
                "Row Selection",
                choices=["rows", "row", "none"],
            ),
            ui.input_action_button("select", "Select 1, 3, 5"),
            ui.input_action_button("clear", "Clear selection"),
        ),
        ui.layout_columns(
            ui.h5("Selected Rows:"),
            ui.output_code("selected_rows"),
            ui.h5("Selected row count:"),
            ui.output_code("selected_row_count"),
            ui.h5("Total row count:"),
            ui.output_code("grid_row_count"),
        ),
        ui.hr(),
        ui.h5("Selected Data Frame"),
        ui.output_data_frame("grid_selected"),
        ui.h5("Data Frame"),
        ui.output_data_frame("grid"),
    )


def make_server(input: Inputs, data: IntoDataFrame):
    @render.data_frame
    def grid():
        return render.DataGrid(
            data=data,
            selection_mode=input.selection_mode(),
            height="300px",
        )

    @render.data_frame
    def grid_selected():
        return render.DataGrid(
            data=grid.data_view(selected=True),
            selection_mode=input.selection_mode(),
        )

    @reactive.effect
    @reactive.event(input.clear)
    async def clear():
        await grid.update_cell_selection(None)

    @reactive.effect
    @reactive.event(input.select)
    async def select_1_3_5():
        await grid.update_cell_selection({"type": "row", "rows": (1, 3, 5)})

    @render.code
    def selected_rows():
        return str(grid.cell_selection()["rows"])

    # Test for selected rows data
    @render.code
    def grid_row_count():
        nrow = grid._nw_data().shape[0]
        return str(nrow)

    @render.code
    def selected_row_count():
        nrow = grid_selected._nw_data().shape[0]
        return str(nrow)


@module.ui
def mod_ui(title: str = "Module"):
    return make_ui(title)


@module.server
def mod_server(
    input: Inputs,
    output: Outputs,
    session: Session,
    data: IntoDataFrame,
):
    make_server(input, data)


app_ui = ui.page_fluid(
    make_ui("global"),
    mod_ui("pandas", "Module - pandas"),
    mod_ui("polars", "Module - polars"),
)


def server(input: Inputs, output: Outputs, session: Session):
    make_server(input, pd_penguins)

    mod_server("pandas", pd_penguins)

    mod_server("polars", pl_penguins)


app = App(app_ui, server)



from shiny import App, ui

app_ui = ui.page_fluid(
    ui.navset_card_underline(
        ui.nav_control(ui.a("Shiny", href="https://shiny.posit.co", target="_blank")),
        ui.nav_control(
            ui.a(
                "Learn Shiny",
                href="https://shiny.posit.co/py/docs/overview.html",
                target="_blank",
            )
        ),
    ),
    id="tab",
)


def server(input, output, session):
    pass


app = App(app_ui, server)



from __future__ import annotations

import sqlite3
from datetime import datetime, timedelta, timezone

import pandas as pd
import plotly.express as px
import scoredata
from plotly_streaming import render_plotly_streaming
from shinywidgets import output_widget

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

THRESHOLD_MID = 0.85
THRESHOLD_MID_COLOR = "rgb(0, 137, 26)"
THRESHOLD_LOW = 0.5
THRESHOLD_LOW_COLOR = "rgb(193, 0, 0)"

# Start a background thread that writes fake data to the SQLite database every second
scoredata.begin()

con = sqlite3.connect(scoredata.SQLITE_DB_URI, uri=True)


def last_modified(con):
    """
    Fast-executing call to get the timestamp of the most recent row in the database.
    We will poll against this in absence of a way to receive a push notification when
    our SQLite database changes.
    """
    return con.execute("select max(timestamp) from accuracy_scores").fetchone()[0]


@reactive.poll(lambda: last_modified(con))
def df():
    """
    @reactive.poll calls a cheap query (`last_modified()`) every 1 second to check if
    the expensive query (`df()`) should be run and downstream calculations should be
    updated.

    By declaring this reactive object at the top-level of the script instead of in the
    server function, all sessions are sharing the same object, so the expensive query is
    only run once no matter how many users are connected.
    """
    tbl = pd.read_sql(
        "select * from accuracy_scores order by timestamp desc, model desc limit ?",
        con,
        params=[150],
    )
    # Convert timestamp to datetime object, which SQLite doesn't support natively
    tbl["timestamp"] = pd.to_datetime(
        tbl["timestamp"], utc=True, format="%Y-%m-%d %H:%M:%S.%f"
    )
    # Create a short label for readability
    tbl["time"] = tbl["timestamp"].dt.strftime("%H:%M:%S")
    # Reverse order of rows
    tbl = tbl.iloc[::-1]

    return tbl


def read_time_period(from_time, to_time):
    tbl = pd.read_sql(
        "select * from accuracy_scores where timestamp between ? and ? order by timestamp, model",
        con,
        params=[from_time, to_time],
    )
    # Treat timestamp as a continuous variable
    tbl["timestamp"] = pd.to_datetime(
        tbl["timestamp"], utc=True, format="%Y-%m-%d %H:%M:%S.%f"
    )
    tbl["time"] = tbl["timestamp"].dt.strftime("%H:%M:%S")

    return tbl


model_names = ["model_1", "model_2", "model_3", "model_4"]
model_colors = {
    name: color
    for name, color in zip(model_names, px.colors.qualitative.D3[0 : len(model_names)])
}


def app_ui(req):
    end_time = datetime.now(timezone.utc)
    start_time = end_time - timedelta(minutes=1)

    return ui.page_sidebar(
        ui.sidebar(
            ui.input_checkbox_group(
                "models", "Models", model_names, selected=model_names
            ),
            ui.input_radio_buttons(
                "timeframe",
                "Timeframe",
                ["Latest", "Specific timeframe"],
                selected="Latest",
            ),
            ui.panel_conditional(
                "input.timeframe === 'Latest'",
                ui.input_selectize(
                    "refresh",
                    "Refresh interval",
                    {
                        0: "Realtime",
                        5: "5 seconds",
                        15: "15 seconds",
                        30: "30 seconds",
                        60 * 5: "5 minutes",
                        60 * 15: "15 minutes",
                    },
                ),
            ),
            ui.panel_conditional(
                "input.timeframe !== 'Latest'",
                ui.input_slider(
                    "timerange",
                    "Time range",
                    min=start_time,
                    max=end_time,
                    value=[start_time, end_time],
                    step=timedelta(seconds=1),
                    time_format="%H:%M:%S",
                ),
            ),
        ),
        ui.div(
            ui.h1("Model monitoring dashboard"),
            ui.p(
                ui.output_ui("value_boxes"),
            ),
            ui.card(output_widget("plot_timeseries")),
            ui.card(output_widget("plot_dist")),
            style="max-width: 800px;",
        ),
        fillable=False,
    )


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.calc
    def recent_df():
        """
        Returns the most recent rows from the database, at the refresh interval
        requested by the user. If the refresh interview is 0, go at maximum speed.
        """
        refresh = int(input.refresh())
        if refresh == 0:
            return df()
        else:
            # This approach works well if you know that input.refresh() is likely to be
            # a longer interval than the underlying changing data source (df()). If not,
            # then this can cause downstream reactives to be invalidated when they
            # didn't need to be.
            reactive.invalidate_later(refresh)
            with reactive.isolate():
                return df()

    @reactive.calc
    def timeframe_df():
        """
        Returns rows from the database within the specified time range. Notice that we
        implement the business logic as a separate function (read_time_period), so it's
        easier to reason about and test.
        """
        start, end = input.timerange()
        return read_time_period(start, end)

    @reactive.calc
    def filtered_df():
        """
        Return the data frame that should be displayed in the app, based on the user's
        input. This will be either the latest rows, or a specific time range. Also
        filter out rows for models that the user has deselected.
        """
        data = recent_df() if input.timeframe() == "Latest" else timeframe_df()

        # Filter the rows so we only include the desired models
        return data[data["model"].isin(input.models())]

    @reactive.calc
    def filtered_model_names():
        return filtered_df()["model"].unique()

    @render.ui
    def value_boxes():
        data = filtered_df()
        models = data["model"].unique().tolist()
        scores_by_model = {
            x: data[data["model"] == x].iloc[-1]["score"] for x in models
        }
        # Round scores to 2 decimal places
        scores_by_model = {x: round(y, 2) for x, y in scores_by_model.items()}

        return ui.layout_column_wrap(
            *[
                # For each model, return a value_box with the score, colored based on
                # how high the score is.
                ui.value_box(
                    model,
                    ui.h2(score),
                    theme=(
                        "text-success"
                        if score > THRESHOLD_MID
                        else "text-warning" if score > THRESHOLD_LOW else "bg-danger"
                    ),
                )
                for model, score in scores_by_model.items()
            ],
            width="135px",
            fixed_width=True,
        )

    @render_plotly_streaming(recreate_key=filtered_model_names, update="data")
    def plot_timeseries():
        """
        Returns a Plotly Figure visualization. Streams new data to the Plotly widget in
        the browser whenever filtered_df() updates, and completely recreates the figure
        when filtered_model_names() changes (see recreate_key=... above).
        """
        fig = px.line(
            filtered_df(),
            x="time",
            y="score",
            labels=dict(score="accuracy"),
            color="model",
            color_discrete_map=model_colors,
            # The default for render_mode is "auto", which switches between
            # type="scatter" and type="scattergl" depending on the number of data
            # points. Switching that value breaks streaming updates, as the type
            # property is read-only. Setting it to "webgl" keeps the type consistent.
            render_mode="webgl",
            template="simple_white",
        )

        fig.add_hline(
            THRESHOLD_LOW,
            line_dash="dash",
            line=dict(color=THRESHOLD_LOW_COLOR, width=2),
            opacity=0.3,
        )
        fig.add_hline(
            THRESHOLD_MID,
            line_dash="dash",
            line=dict(color=THRESHOLD_MID_COLOR, width=2),
            opacity=0.3,
        )

        fig.update_yaxes(range=[0, 1], fixedrange=True)
        fig.update_xaxes(fixedrange=True, tickangle=60)

        return fig

    @render_plotly_streaming(recreate_key=filtered_model_names, update="data")
    def plot_dist():
        fig = px.histogram(
            filtered_df(),
            facet_row="model",
            nbins=20,
            x="score",
            labels=dict(score="accuracy"),
            color="model",
            color_discrete_map=model_colors,
            template="simple_white",
        )

        fig.add_vline(
            THRESHOLD_LOW,
            line_dash="dash",
            line=dict(color=THRESHOLD_LOW_COLOR, width=2),
            opacity=0.3,
        )
        fig.add_vline(
            THRESHOLD_MID,
            line_dash="dash",
            line=dict(color=THRESHOLD_MID_COLOR, width=2),
            opacity=0.3,
        )

        # From https://plotly.com/python/facet-plots/#customizing-subplot-figure-titles
        fig.for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))

        fig.update_yaxes(matches=None)
        fig.update_xaxes(range=[0, 1], fixedrange=True)
        fig.layout.height = 500

        return fig

    @reactive.effect
    def update_time_range():
        """
        Every 5 seconds, update the custom time range slider's min and max values to
        reflect the current min and max values in the database.
        """

        reactive.invalidate_later(15)
        try:
            min_time, max_time = pd.to_datetime(
                con.execute(
                    "select min(timestamp), max(timestamp) from accuracy_scores"
                ).fetchone(),
                utc=True,
            )
            ui.update_slider(
                "timerange",
                min=min_time.replace(tzinfo=timezone.utc),
                max=max_time.replace(tzinfo=timezone.utc),
            )
        except sqlite3.OperationalError:
            # Sometimes this executes before the background thread has had a
            # chance to even create the sample data table. In that case, just
            # ignore the error and try again later.
            pass


app = App(app_ui, server)



from shiny import reactive, req
from shiny.express import input, render, ui
from shiny.types import SafeException

ui.input_action_button("safe", "Throw a safe error")


@render.ui
def safe():
    # This error _won't_ be sanitized when deployed (i.e., it's "safe")
    raise SafeException(f"You've clicked {str(safe_click())} times")


ui.input_action_button("unsafe", "Throw an unsafe error")


@render.ui
def unsafe():
    req(input.unsafe())
    # This error _will_ be sanitized when deployed (i.e., it's "unsafe")
    raise Exception(f"Super secret number of clicks: {str(input.unsafe())}")


ui.input_text(
    "txt",
    "Enter some text below, then remove it. Notice how the text is never fully removed.",
)


@render.ui
def txt_out():
    req(input.txt(), cancel_output=True)
    return input.txt()


@reactive.calc
def safe_click():
    req(input.safe())
    return input.safe()


@reactive.effect
def _():
    req(input.unsafe())
    print("unsafe clicks:", input.unsafe())
    # raise Exception("Observer exception: this should cause a crash")



import pandas as pd
import polars as pl
import seaborn as sns
from narwhals.stable.v1.typing import IntoDataFrame

from shiny import App, Inputs, Outputs, Session, module, reactive, render, ui

df = pd.DataFrame(
    sns.load_dataset(  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
        "iris"
    )  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
)

distinct_df = df.drop_duplicates(subset=["species"])
idxs = [0, 1, 50, 51, 100, 101]
distinct_df = df.iloc[idxs]


@module.ui
def mod_ui():
    return ui.TagList(
        ui.h2("Iris Dataset"),
        ui.layout_column_wrap(
            ui.input_action_button("reset_df", "Reset Dataframe"),
            ui.input_action_button("update_sort", "Update sort"),
            ui.input_action_button("update_filter", "Update filter"),
            width=1 / 3,
        ),
        ui.br(),
        ui.output_data_frame("iris_df"),
        ui.h2("Data view indices"),
        ui.output_text_verbatim("data_view_rows"),
        ui.h2("Indices when view_selected=True"),
        ui.output_text_verbatim("data_view_selected_true"),
        ui.h2("Indices when view_selected=False"),
        ui.output_text_verbatim("data_view_selected_false"),
        ui.h2("Show selected cell"),
        ui.output_text_verbatim("cell_selection"),
        ui.h2("Type:"),
        ui.output_text_verbatim("df_type"),
    )


@module.server
def mod_server(
    input: Inputs,
    output: Outputs,
    session: Session,
    data: IntoDataFrame,
):
    @render.text
    def df_type():
        return str(type(data))

    @render.data_frame
    def iris_df():
        return render.DataGrid(
            data=data,
            filters=True,
            selection_mode="rows",
        )

    @render.code
    def data_view_rows():
        return str(iris_df.data_view_rows())

    @render.code
    def data_view_selected_false():  # pyright: ignore[reportUnknownParameterType]
        # TODO-barret; remove type check when DataFrameT is implemented
        data_view_all = iris_df.data_view(selected=False)

        if isinstance(data_view_all, pd.DataFrame):
            return str(
                [
                    int(val)
                    for val in data_view_all.index.values  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType, reportUnknownArgumentType]
                ]
            )
        elif isinstance(data_view_all, pl.DataFrame):
            return str(data_view_all.get_column("index").to_list())
        else:
            raise ValueError("Invalid data type")

    @render.code
    def data_view_selected_true():  # pyright: ignore[reportUnknownParameterType]
        # TODO-barret; remove type check when DataFrameT is implemented
        data_view_selected = iris_df.data_view(selected=True)

        if isinstance(data_view_selected, pd.DataFrame):
            return str(
                [
                    int(val)
                    for val in data_view_selected.index.values  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType, reportUnknownArgumentType]
                ]
            )
        elif isinstance(data_view_selected, pl.DataFrame):
            return str(data_view_selected["index"].to_list())
        else:
            raise ValueError("Invalid data type")

    @render.code
    def cell_selection():  # pyright: ignore[reportUnknownParameterType]
        return str(iris_df.cell_selection()["rows"])

    @reactive.effect
    @reactive.event(input.reset_df)
    def reset_df():
        iris_df._reset_reactives()

    @reactive.effect
    @reactive.event(input.update_sort)
    async def update_sort():
        await iris_df.update_sort([1, {"col": 2, "desc": False}])

    @reactive.effect
    @reactive.event(input.update_filter)
    async def update_filter():
        await iris_df.update_filter(
            [{"col": 0, "value": [5, 6.9]}, {"col": 4, "value": "v"}]
        )


app_ui = ui.page_fluid(
    ui.navset_card_underline(
        ui.nav_panel(
            "pandas",
            mod_ui("pandas"),
        ),
        ui.nav_panel(
            "polars",
            mod_ui("polars"),
        ),
        id="tab",
    ),
)


def server(input: Inputs, output: Outputs, session: Session) -> None:
    mod_server("pandas", distinct_df)
    mod_server(
        "polars", pl.from_pandas(distinct_df).with_columns(pl.Series("index", idxs))
    )


app = App(app_ui, server)



from __future__ import annotations

__all__ = (
    "output_plot",
    "output_image",
    "output_text",
    "output_code",
    "output_text_verbatim",
    "output_table",
    "output_ui",
)

from typing import Optional

from htmltools import Tag, TagAttrValue, TagFunction, css, div, tags

from .._docstring import add_example, no_example
from .._namespaces import resolve_id
from ..types import MISSING, MISSING_TYPE
from ._plot_output_opts import (
    BrushOpts,
    ClickOpts,
    DblClickOpts,
    HoverOpts,
    brush_opts,
    click_opts,
    dblclick_opts,
    format_opt_names,
    hover_opts,
)
from .fill import as_fill_item, as_fillable_container


@add_example()
def output_plot(
    id: str,
    width: str | float | int = "100%",
    height: str | float | int = "400px",
    *,
    inline: bool = False,
    click: bool | ClickOpts = False,
    dblclick: bool | DblClickOpts = False,
    hover: bool | HoverOpts = False,
    brush: bool | BrushOpts = False,
    fill: bool | MISSING_TYPE = MISSING,
) -> Tag:
    """
    Create a output container for a static plot.

    Place a :class:`~shiny.render.plot` result in the user interface. See
    :class:`~shiny.render.plot` for more details on what types of plots are supported.

    Parameters
    ----------
    id
        An output id.
    width
        The CSS width, e.g. '400px', or '100%'.
    height
        The CSS height, e.g. '100%' or '600px'.
    inline
        If ``True``, the result is displayed inline.
    click
        This can be a boolean or an object created by :func:`~shiny.ui.click_opts`. The
        default is `False`, but if you use `True` (or equivalently, `click_opts()`), the
        plot will send coordinates to the server whenever it is clicked, and the value
        will be accessible via `input.xx_click()`, where `xx` is replaced with the ID of
        this plot. The input value will be a dictionary with `x` and `y` elements
        indicating the mouse position.
    dblclick
        This is just like the `click` parameter, but for double-click events.
    hover
        Similar to the `click` argument, this can be a boolean or an object created by
        :func:`~shiny.ui.hover_opts`. The default is `False`, but if you use `True` (or
        equivalently, `hover_opts()`), the plot will send coordinates to the server
        whenever it is clicked, and the value will be accessible via `input.xx_hover()`,
        where `xx` is replaced with the ID of this plot. The input value will be a
        dictionary with `x` and `y` elements indicating the mouse position. To control
        the hover time or hover delay type, use :func:`~shiny.ui.hover_opts`.
    brush
        Similar to the `click` argument, this can be a boolean or an object created by
        :func:`~shiny.ui.brush_opts`. The default is `False`, but if you use `True` (or
        equivalently, `brush_opts()`), the plot will allow the user to "brush" in the
        plotting area, and will send information about the brushed area to the server,
        and the value will be accessible via `input.plot_brush()`. Brushing means that
        the user will be able to draw a rectangle in the plotting area and drag it
        around. The value will be a named list with `xmin`, `xmax`, `ymin`, and `ymax`
        elements indicating the brush area. To control the brush behavior, use
        :func:`~shiny.ui.brush_opts`. Multiple `output_image`/`output_plot` calls may
        share the same `id` value; brushing one image or plot will cause any other
        brushes with the same `id` to disappear.
    fill
        Whether or not to allow the plot output to grow/shrink to fit a fillable
        container with an opinionated height (e.g., :func:`~shiny.ui.page_fillable`). If
        no `fill` value is provided, it will default to the inverse of `inline`.

    Returns
    -------
    :
        A UI element

    See Also
    --------
    * :class:`~shiny.render.plot`
    * :func:`~shiny.ui.output_image`
    """
    if isinstance(fill, MISSING_TYPE):
        fill = not inline
    res = output_image(
        id=id,
        width=width,
        height=height,
        inline=inline,
        click=click,
        dblclick=dblclick,
        hover=hover,
        brush=brush,
        fill=fill,
    )
    res.add_class("shiny-plot-output")
    return res


@add_example()
def output_image(
    id: str,
    width: str | float | int = "100%",
    height: str | float | int = "400px",
    *,
    inline: bool = False,
    click: bool | ClickOpts = False,
    dblclick: bool | DblClickOpts = False,
    hover: bool | HoverOpts = False,
    brush: bool | BrushOpts = False,
    fill: bool = False,
) -> Tag:
    """
    Create a output container for a static image.

    Parameters
    ----------
    id
        An output id.
    width
        The CSS width, e.g. '400px', or '100%'.
    height
        The CSS height, e.g. '100%' or '600px'.
    inline
        If ``True``, the result is displayed inline.
    click
        This can be a boolean or an object created by :func:`~shiny.ui.click_opts`. The
        default is `False`, but if you use `True` (or equivalently, `click_opts()`), the
        plot will send coordinates to the server whenever it is clicked, and the value
        will be accessible via `input.xx_click()`, where `xx` is replaced with the ID of
        this plot. The input value will be a dictionary with `x` and `y` elements
        indicating the mouse position.
    dblclick
        This is just like the `click` parameter, but for double-click events.
    hover
        Similar to the `click` argument, this can be a boolean or an object created by
        :func:`~shiny.ui.hover_opts`. The default is `False`, but if you use `True` (or
        equivalently, `hover_opts()`), the plot will send coordinates to the server
        whenever it is clicked, and the value will be accessible via `input.xx_hover()`,
        where `xx` is replaced with the ID of this plot. The input value will be a
        dictionary with `x` and `y` elements indicating the mouse position. To control
        the hover time or hover delay type, use :func:`~shiny.ui.hover_opts`.
    brush
        Similar to the `click` argument, this can be a boolean or an object created by
        :func:`~shiny.ui.brush_opts`. The default is `False`, but if you use `True` (or
        equivalently, `brush_opts()`), the plot will allow the user to "brush" in the
        plotting area, and will send information about the brushed area to the server,
        and the value will be accessible via `input.plot_brush()`. Brushing means that
        the user will be able to draw a rectangle in the plotting area and drag it
        around. The value will be a named list with `xmin`, `xmax`, `ymin`, and `ymax`
        elements indicating the brush area. To control the brush behavior, use
        :func:`~shiny.ui.brush_opts`. Multiple `output_image`/`output_plot` calls may
        share the same `id` value; brushing one image or plot will cause any other
        brushes with the same `id` to disappear.
    fill
        Whether or not to allow the image output to grow/shrink to fit a fillable
        container with an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).

    Returns
    -------
    :
        A UI element

    See Also
    --------
    * :class:`~shiny.render.image`
    * :func:`~shiny.ui.output_plot`
    """
    func = tags.span if inline else div
    width = f"{width}px" if isinstance(width, (float, int)) else width
    height = f"{height}px" if isinstance(height, (float, int)) else height
    style = None if inline else css(width=width, height=height)

    args: dict[str, str] = dict()

    id_resolved = resolve_id(id)

    if click is not False:
        if click is True:
            click = click_opts()
        click["id"] = id_resolved + "_click"
        args.update(**format_opt_names(click, "click"))

    if dblclick is not False:
        if dblclick is True:
            dblclick = dblclick_opts()
        dblclick["id"] = id_resolved + "_dblclick"
        args.update(**format_opt_names(dblclick, "dblclick"))

    if hover is not False:
        if hover is True:
            hover = hover_opts()
        hover["id"] = id_resolved + "_hover"
        args.update(**format_opt_names(hover, "hover"))

    if brush is not False:
        if brush is True:
            brush = brush_opts()
        brush["id"] = id_resolved + "_brush"
        args.update(**format_opt_names(brush, "brush"))

    container = func(
        id=id_resolved,
        class_="shiny-image-output",
        style=style,
        **args,
    )
    if fill:
        container = as_fill_item(container)

    return container


@add_example()
def output_text(
    id: str, inline: bool = False, container: Optional[TagFunction] = None
) -> Tag:
    """
    Create a output container for some text.

    Parameters
    ----------
    id
        An output id.
    inline
        If ``True``, the result is displayed inline.
    container
        A Callable that returns the output container.

    Returns
    -------
    :
        A UI element

    Note
    ----
    Text is HTML-escaped prior to rendering.

    See Also
    --------
    * :class:`~shiny.render.text`
    * :func:`~shiny.ui.output_text_verbatim`
    """

    if not container:
        container = tags.span if inline else tags.div
    return container(id=resolve_id(id), class_="shiny-text-output")


@no_example()
def output_code(id: str, placeholder: bool = True) -> Tag:
    """
    Create a output container for code (monospaced text).

    This is similar to :func:`~shiny.ui.output_text`, except that it displays the text
    in a fixed-width container with a gray-ish background color and border.

    Parameters
    ----------
    id
        An output id.
    placeholder
        If the output is empty or ``None``, should an empty rectangle be displayed to
        serve as a placeholder? (This does not affect behavior when the output is
        nonempty.)

    Returns
    -------
    :
        A UI element

    Note
    ----
    This function is currently the same as :func:`~shiny.ui.output_text_verbatim`, but
    this may change in future versions of Shiny.

    See Also
    --------
    * :class:`~shiny.render.text`
    * :func:`~shiny.ui.output_text`
    * :func:`~shiny.ui.output_text_verbatim`

    Example
    -------
    See :func:`~shiny.ui.output_text`
    """

    cls = "shiny-text-output" + (" noplaceholder" if not placeholder else "")
    return tags.pre(id=resolve_id(id), class_=cls)


@add_example(ex_dir="../api-examples/input_text")
def output_text_verbatim(id: str, placeholder: bool = False) -> Tag:
    """
    Create a output container for some text.

    Place a :class:`~shiny.render.text` result in the user interface.
    Differs from :func:`~shiny.ui.output_text` in that it wraps the text in a
    fixed-width container with a gray-ish background color and border.

    Parameters
    ----------
    id
        An output id.
    placeholder
        If the output is empty or ``None``, should an empty rectangle be displayed to
        serve as a placeholder? (This does not affect behavior when the output
        is nonempty.)

    Returns
    -------
    :
        A UI element

    See Also
    --------
    * :class:`~shiny.render.text`
    * :func:`~shiny.ui.output_text`

    Example
    -------
    See :func:`~shiny.ui.output_text`
    """

    cls = "shiny-text-output" + (" noplaceholder" if not placeholder else "")
    return tags.pre(id=resolve_id(id), class_=cls)


@add_example()
def output_table(id: str, **kwargs: TagAttrValue) -> Tag:
    """
    Create a output container for a table.

    Parameters
    ----------
    id
        An output id.
    **kwargs
        Additional attributes to add to the container.

    Returns
    -------
    :

    See Also
    --------
    * :class:`~shiny.render.table`
    """
    return tags.div({"class": "shiny-html-output"}, id=resolve_id(id), **kwargs)


@add_example()
def output_ui(
    id: str,
    inline: bool = False,
    container: Optional[TagFunction] = None,
    fill: bool = False,
    fillable: bool = False,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a output container for a UI (i.e., HTML) element.

    Parameters
    ----------
    id
        An output id.
    inline
        If ``True``, the result is displayed inline.
    container
        A Callable that returns the output container.
    fill
        Whether or not to allow the UI output to grow/shrink to fit a fillable container
        with an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).
    fillable
        Whether or not the UI output area should be considered a fillable (i.e.,
        flexbox) container.
    **kwargs
        Attributes to be applied to the output container.

    Returns
    -------
    :
        A UI element

    See Also
    --------
    * :class:`~shiny.render.ui`
    * :func:`~shiny.ui.output_text`
    """

    if not container:
        container = tags.span if inline else tags.div
    res = container({"class": "shiny-html-output"}, id=resolve_id(id), **kwargs)
    if fill:
        res = as_fill_item(res)
    if fillable:
        res = as_fillable_container(res)
    return res



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_page_navbar_selected(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    page_navbar_selected = controller.PageNavbar(page, "page_navbar_selected")
    page_navbar_selected.expect_value("View")
    page_navbar_selected._expect_content_text(
        "This page could be used to view the dataset."
    )



import aiohttp
from bs4 import BeautifulSoup

recipe_prompt = """
You are RecipeExtractorGPT.
Your goal is to extract recipe content from text and return a JSON representation of the useful information.

The JSON should be structured like this:

```
{
  "title": "Scrambled eggs",
  "ingredients": {
    "eggs": "2",
    "butter": "1 tbsp",
    "milk": "1 tbsp",
    "salt": "1 pinch"
  },
  "directions": [
    "Beat eggs, milk, and salt together in a bowl until thoroughly combined.",
    "Heat butter in a large skillet over medium-high heat. Pour egg mixture into the hot skillet; cook and stir until eggs are set, 3 to 5 minutes."
  ],
  "servings": 2,
  "prep_time": 5,
  "cook_time": 5,
  "total_time": 10,
  "tags": [
    "breakfast",
    "eggs",
    "scrambled"
  ],
  "source": "https://recipes.com/scrambled-eggs/",
}
```

The user will provide text content from a web page.
It is not very well structured, but the recipe is in there.
Please look carefully for the useful information about the recipe.
IMPORTANT: Return the result as JSON in a Markdown code block surrounded with three backticks!
"""


async def scrape_page_with_url(url: str, max_length: int = 14000) -> str:
    """
    Given a URL, scrapes the web page and return the contents. This also adds adds the
    URL to the beginning of the text.

    Parameters
    ----------
    url:
      The URL to scrape
    max_length:
      Max length of recipe text to process. This is to prevent the model from running
      out of tokens. 14000 bytes translates to approximately 3200 tokens.
    """
    contents = await scrape_page(url)
    # Trim the string so that the prompt and reply will fit in the token limit.. It
    # would be better to trim by tokens, but that requires using the tiktoken package,
    # which can be very slow to load when running on containerized servers, because it
    # needs to download the model from the internet each time the container starts.
    contents = contents[:max_length]
    return f"From: {url}\n\n" + contents


async def scrape_page(url: str) -> str:
    # Asynchronously send an HTTP request to the URL.
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            if response.status != 200:
                raise aiohttp.ClientError(f"An error occurred: {response.status}")
            html = await response.text()

    # Parse the HTML content using BeautifulSoup
    soup = BeautifulSoup(html, "html.parser")

    # Remove script and style elements
    for script in soup(["script", "style"]):
        script.decompose()

    # List of element IDs or class names to remove
    elements_to_remove = [
        "header",
        "footer",
        "sidebar",
        "nav",
        "menu",
        "ad",
        "advertisement",
        "cookie-banner",
        "popup",
        "social",
        "breadcrumb",
        "pagination",
        "comment",
        "comments",
    ]

    # Remove unwanted elements by ID or class name
    for element in elements_to_remove:
        for e in soup.find_all(id=element) + soup.find_all(class_=element):
            e.decompose()

    # Extract text from the remaining HTML tags
    text = " ".join(soup.stripped_strings)

    return text



from datetime import date

from shiny import App, Inputs, Outputs, Session, ui

app_ui = ui.page_fluid(
    # Default start and end is the current date in the client's time zone
    ui.input_date_range("daterange1", "Date range:"),
    # Set start and end dates
    ui.input_date_range(
        "daterange2", "Set start and end date:", start="2001-01-01", end="2010-12-31"
    ),
    # Start and end are always specified in yyyy-mm-dd, even if the display
    # format is different
    ui.input_date_range(
        "daterange3",
        "Min, max, start, and end dates are set with custom format and separator:",
        start="2001-01-01",
        end="2010-12-31",
        min="2001-01-01",
        max="2012-12-21",
        format="mm/dd/yy",
        separator=" - ",
    ),
    # Pass in Date objects
    ui.input_date_range(
        "daterange4",
        "Default start and end use date objects:",
        start=date(2001, 1, 1),
        end=date(2010, 12, 31),
    ),
    # Use different language and different first day of week
    ui.input_date_range(
        "daterange5",
        "Language is German and we starts on Monday:",
        language="de",
        weekstart=1,
    ),
    # Start with decade view instead of default month view
    ui.input_date_range(
        "daterange6", "Start Date picker in Decade view:", startview="decade"
    ),
    # Set width of the daterange field
    ui.input_date_range("daterange7", "Set width of text input:", width="600px"),
    # Set autoclose to false
    ui.input_date_range("daterange8", "Auto close is disabled:", autoclose=False),
)


def server(input: Inputs, output: Outputs, session: Session):
    pass


app = App(app_ui, server)



from __future__ import annotations

import os
from typing import Any, Dict, List, Optional, cast, overload

from htmltools import (
    HTMLDependency,
    Tag,
    TagChild,
    TagList,
    TagNode,
    head_content,
    tags,
)

from .._typing_extensions import TypeGuard
from ..session import Session, require_active_session
from ..types import MISSING, MISSING_TYPE


def shiny_input_label(id: str, label: TagChild = None) -> Tag:
    cls = "control-label" + ("" if label else " shiny-label-null")
    return tags.label(label, class_=cls, id=id + "-label", for_=id)


@overload
def get_window_title(
    title: None,
    window_title: MISSING_TYPE,
) -> None: ...


@overload
def get_window_title(
    title: None,
    window_title: str,
) -> HTMLDependency: ...


@overload
def get_window_title(
    title: str | Tag | TagList,
    window_title: str | MISSING_TYPE,
) -> HTMLDependency: ...


def get_window_title(
    title: Optional[str | Tag | TagList],
    window_title: str | MISSING_TYPE = MISSING,
) -> Optional[HTMLDependency]:
    if title is not None and isinstance(window_title, MISSING_TYPE):
        window_title = _find_child_strings(title)

    if isinstance(window_title, MISSING_TYPE):
        return None
    else:
        return head_content(tags.title(window_title))


def _find_child_strings(x: TagList | TagNode) -> str:
    if isinstance(x, Tag) and x.name not in ("script", "style"):
        x = x.children
    if isinstance(x, TagList):
        strings = [_find_child_strings(y) for y in x]
        return " ".join(filter(lambda x: x != "", strings))
    if isinstance(x, str):
        return x
    return ""


def _session_on_flush_send_msg(
    id: str, session: Session | None, msg: dict[str, object]
) -> None:
    session = require_active_session(session)
    session.on_flush(lambda: session.send_input_message(id, msg), once=True)


def is_01_scalar(x: object) -> TypeGuard[float]:
    return isinstance(x, (int, float)) and x >= 0.0 and x <= 1.0


def css_no_sub(**kwargs: str | float | None) -> Optional[str]:
    """
    Altered from py-htmltools's `css()`. Does not support substitutions of any kind.
    """
    res = ""
    for k, v in kwargs.items():
        if v is None:
            continue
        v = " ".join(v) if isinstance(v, list) else str(v)
        res += k + ":" + v + ";"
    return None if res == "" else res


class JSEval(str):
    pass


def js_eval(x: str) -> JSEval:
    """
    Mark a function as a JavaScript evaluation.
    Some components like `input_selectize` allow you to send JavaScript functions
    to the client.
    This function marks a string as a JavaScript evaluation so that it can be properly
    sent to the client library.
    """
    return JSEval(x)


def extract_js_keys(options: Dict[str, Any], parent_key: str = "") -> List[str]:
    """
    This function extracts JavaScript (JS) and HTML objects from the provided dictionary.
    This is used to identify which options should be evaluated by the client.

    Parameters:
    options (Dict[str, Any]): A dictionary containing various options and pulls out the keys
    of those options which are of type JSEval.

    Returns:
    A list of keys which identify the JSEval options
    """

    # TODO This only works on nested dictionaries, and we may need to extend it to
    # recurse through lists as well.
    js_html_keys: List[str] = []
    for key, value in options.items():
        full_key = f"{parent_key}.{key}" if parent_key else key
        if isinstance(value, JSEval):
            js_html_keys.append(full_key)
        elif isinstance(value, dict):
            value = cast(Dict[str, Any], value)
            js_html_keys.extend(extract_js_keys(value, full_key))
    return js_html_keys


def path_pkg_www(*args: str) -> str:
    base = os.path.join(os.path.dirname(__file__), "..", "www", "shared")
    return os.path.join(base, *args)
